/*
 * Echo Server Example: Server Module.
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <pthread.h>


#define BUSYFMT	"Sorry, Echo Server is currently handling %d connections..." \
		"\n\r\tand it feels too busy.\n\r\n\rTry again later\n\r\n\r"


struct childarg {
	int fd;
	struct sockaddr_in addr;
};


static pthread_mutex_t print_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t conns_mutex = PTHREAD_MUTEX_INITIALIZER;
static int conns = 0;


static void
server_print(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	pthread_mutex_lock(&print_mutex);
	vprintf(fmt, ap);
	pthread_mutex_unlock(&print_mutex);
	va_end(ap);
}


static int
readline(int cfd, char *buf, int buflen)
{
	int r, len;

	r = 0;

	while (buflen--) {
		len = read(cfd, buf, 1);
		if (len <= 0) {
			if (len == 0 || errno == ECONNRESET) {
				break;
			}

			fprintf(stderr, "Can't read: %s\n", strerror(errno));
			exit(-1);
			/* NOTREACHED */
		}

		r++;
		buf++;

		if (*(buf - 1) == '\n') {
			break;
		}

	}

	*buf = '\0';

	return r;
}


static void *
child_thread(void *arg)
{
	struct childarg *ca;
	in_addr_t addr;

	char buf[1024];
	int len;

	ca = (struct childarg *)arg;

	snprintf(buf, sizeof(buf),
	    "Greetings... I'm echo server, id=%p, you are connection %d\n"
	    "Type 'quit' to exit from this wonderful chat program\n\n",
	    pthread_self(), conns);

	write(ca->fd, buf, strlen(buf));

	while (1) {
		/*
		 * Read line.
		 */
		len = readline(ca->fd, buf, sizeof(buf));
		if (len) {
			/*
			 * Echo line.
			 */
			write(ca->fd, buf, len);
			if (strncmp(buf, "quit\n", 4)) {
				continue;
			}
		}

		/*
		 * Broken pipe or 'quit' command.
		 */
		addr = ntohl(ca->addr.sin_addr.s_addr);
		server_print("Connection with %d.%d.%d.%d:%d terminated\n",
		    (addr >> 24) & 0xFF,
		    (addr >> 16) & 0xFF,
		    (addr >>  8) & 0xFF,
		    (addr >>  0) & 0xFF,
		    ntohs(ca->addr.sin_port));

		close(ca->fd);

		pthread_mutex_lock(&conns_mutex);
		conns--;
		pthread_mutex_unlock(&conns_mutex);

		free(ca);

		pthread_exit(0);
	}

	return NULL;
}


/*
 * The main thread listens.
 * There is one child thread for each client connection.
 */
void PLEXPORT
start_server(uint16_t port, int max)
{
	struct sockaddr_in addr;
	struct childarg *ca;
	socklen_t caddr_len;
	pthread_t child;
	in_addr_t caddr;
	int lfd, e;

	lfd = socket(AF_INET, SOCK_STREAM, 0);
	if (lfd < 0) {
		fprintf(stderr, "Can't socket: %s\n", strerror(errno));
		exit(-1);
		/* NOTREACHED */
	}

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(port);

	e = SO_REUSEADDR;
	if (setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR, (void *)&e, sizeof(e))
	    < 0) {
		fprintf(stderr, "Can't setsockopt: %s\n", strerror(errno));
		exit(-1);
		/* NOTREACHED */
	}

	if (bind(lfd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		fprintf(stderr, "Can't bind: %s\n", strerror(errno));
		exit(-1);
		/* NOTREACHED */
	}

	if (listen(lfd, 5) < 0) {
		fprintf(stderr, "Can't listen: %s\n", strerror(errno));
		exit(-1);
		/* NOTREACHED */
	}

	ca = NULL;

	while (1) {
		if (ca == NULL) {
			ca = (struct childarg *)malloc(sizeof(*ca));
			if (ca == NULL) {
				fprintf(stderr, "Can't malloc: %s\n",
				    strerror(errno));
				exit(-1);
				/* NOTREACHED */
			}
		}

		bzero(ca, sizeof(*ca));
		caddr_len = sizeof(ca->addr);
		ca->fd = accept(lfd, (struct sockaddr *)&ca->addr, &caddr_len);
		if (ca->fd < 0) {
			if (errno == EINTR) {
				continue;
			}

			fprintf(stderr, "Can't accept: %s\n", strerror(errno));
			exit(-1);
			/* NOTREACHED */
		}

		pthread_mutex_lock(&conns_mutex);

		if (conns >= max) {
			char buf[256];

			snprintf(buf, sizeof(buf), BUSYFMT, conns);
			pthread_mutex_unlock(&conns_mutex);
			write(ca->fd, buf, strlen(buf));
			close(ca->fd);
			continue;
		}

		conns++;

		caddr = ntohl(ca->addr.sin_addr.s_addr);

		server_print("New connection %d from %d.%d.%d.%d:%d\n",
		    conns,
		    (caddr >> 24) & 0xFF,
		    (caddr >> 16) & 0xFF,
		    (caddr >>  8) & 0xFF,
		    (caddr >>  0) & 0xFF,
		    ntohs(ca->addr.sin_port));

		pthread_mutex_unlock(&conns_mutex);
	    
		e = pthread_create(&child, NULL, (void *)child_thread, ca);
		if (e != 0) {
			fprintf(stderr, "Can't pthread_create: %s\n",
			    strerror(e));
			exit (-1);
			/* NOTREACHED */
		}

		ca = NULL;
	}
}


