/**
 * @file src/core/boot/boot.c
 * Peludo Bootstrap Code.
 */
#ifndef PKERNEL_CODE_SIZE
#error "PKERNEL_CODE_SIZE must be defined."
#endif

#ifndef PKERNEL_DATA_SIZE
#error "PKERNEL_DATA_SIZE must be defined."
#endif

#ifndef PKERNEL_GOT_SIZE
#error "PKERNEL_GOT_SIZE must be defined."
#endif

#include <plc.h>

extern long boot_end();		/* See boot.lds */


/**
 * Entry point of bootstrap code.
 * @param argc Number of arguments;
 * @param argv Array of argument strings.
 * @return A Peludo Status code or the program's exit value.
 * @see PLSTATUS
 */
long __attribute__((section(".boot")))
boot(int argc, char **argv)
{
	long kernel, pld, addr, kdata, kdatasiz, i, err, st;

	/*
	 * Initialize the low level PLC api.
	 */
	if ((st = plsys_init()) != PLSTATUS_SUCCESS) {
		return st;
	}

	/* Kernel buffer starts at the end of bootstrap */
	addr = plsys_curpage() + (long)boot_end;

	kdatasiz = PKERNEL_DATA_SIZE + PKERNEL_GOT_SIZE;

	st = _plsys_palloc(&err, &kernel,
	    PLALIGN(PKERNEL_CODE_SIZE, PLPAGE_SIZE) +
	    PLALIGN(kdatasiz, PLPAGE_SIZE),
	    PLMFLAGS_R | PLMFLAGS_W);

	if (st != PLSTATUS_SUCCESS) {
		return st;
	}

	/*
	 * Copy the kernel code.
	 */
	plsys_memmove((void *)kernel, (void *)addr, PKERNEL_CODE_SIZE);

	if (PKERNEL_DATA_SIZE) {

		kdata = kernel + PLALIGN(PKERNEL_CODE_SIZE, PLPAGE_SIZE);

		/*
		 * Copy the kernel data.
		 */
		plsys_memmove((void *)kdata,
		    (void *)addr + PKERNEL_CODE_SIZE, kdatasiz);

		/*
		 * Reloc the GOT (if any).
		 */
		for (i = 0, kdata += PKERNEL_DATA_SIZE;
		    i < PKERNEL_GOT_SIZE; i += sizeof(long)) {
			*((long *)(kdata + i)) += kernel;
		}
	}

	/*
	 * Kernel code section is executable.
	 */
	st = _plsys_mprotect(&err, kernel, PKERNEL_CODE_SIZE,
	    PLMFLAGS_R | PLMFLAGS_X);
	if (st == PLSTATUS_SUCCESS) {
		/*
		 * PLD files are located after the kernel.
		 */
		pld = addr + PKERNEL_CODE_SIZE + PKERNEL_DATA_SIZE +
		    PKERNEL_GOT_SIZE;

		/*
		 * Start the kernel.
		 */
		st = ((long(*)(int, char **, long))kernel)(argc, argv, pld);
	}

	(void)_plsys_pfree(&err, kernel,
	    PLALIGN(PKERNEL_CODE_SIZE, PLPAGE_SIZE) +
	    PLALIGN(PKERNEL_DATA_SIZE + PKERNEL_GOT_SIZE, PLPAGE_SIZE));

	plsys_release();

	return st;
}


