/**
 * @file src/include/platomic.h
 * Peludo mini C library - Atomic operations.
 */
#ifndef PLATOMIC_H
#define PLATOMIC_H


long platomic_dec_positive(long *);


/**
 * Atomically increments a value by one.
 * @param value Value to increment.
 * @see platomic_dec()
 */
static long platomic_inc(long *value) __attribute__((always_inline));


/**
 * Atomically decrements a value by one.
 * @param value Value to decrement.
 * @see platomic_inc()
 */
static long platomic_dec(long *value) __attribute__((always_inline));


/**
 * Atomic Bit Test and Set.
 * @param addr Value to test and set.
 * @param bit  Bit number to test and set.
 * @return The previous state of the specified bit: A non-zero value
 * means ON, a zero value means off.
 */
static long platomic_bts(long *addr, int bit) __attribute__((always_inline));


/**
 * Interlocked Exchange.
 * <p>Atomically compare cval against the value pointed and if they are equal
 * the value pointed by addr is updated to nval.</p>
 * @param addr Value to exchange;
 * @param cval Value the addr variable should point to in order
 * to do the change.
 * @param nval Value to put into addr when exchange is done.
 * @return If the exchange is done cval is returned; otherwise the current
 * value pointed by addr is returned.
 */
static long platomic_iex(long *addr, long cval, long nval)
					__attribute__((always_inline));


/*
 * Implementations.
 */
#ifdef PLARCH_i386
static inline long
platomic_inc(long *value)
{
	volatile long ret;

	__asm__ __volatile__(
	    "movl %1, %%eax	\n"
	    "lock incl (%%eax)	\n"
	    "movl (%%eax), %%eax	\n"
	    : "=r" (ret) : "m" (value));

	return ret;
}


static inline long
platomic_dec(long *value)
{
	volatile long ret;

	__asm__ __volatile__(
	    "movl %1, %%eax	\n"
	    "lock decl (%%eax)	\n"
	    "movl (%%eax), %%eax\n"
	    : "=r" (ret) : "m" (value));

	return ret;
}


static inline long
platomic_bts(long *addr, int bit)
{
	long ret = 0;

	__asm__ __volatile__(
	    "xorl %%eax, %%eax	\n"
	    "lock btsl %2, %1	\n"
	    "setc %%al		\n"
	    "movl %%eax, %0	\n"
	    :: "g" (ret), "m" (*(addr)), "ir" (bit));

	return ret;
}


static inline long
platomic_iex(long *addr, long cval, long nval)
{
	long ret = 0;

	__asm__ __volatile__(
	    "pushl %%ebx			\n"
	    "movl  %1, %%eax			\n"
	    "movl  %2, %%ebx			\n"
	    "movl  %3, %%edx			\n"
	    "lock cmpxchgl %%ebx, (%%edx)	\n"
	    "popl  %%ebx			\n"
	    "movl  %%eax, %0			\n"
	    :: "g" (ret), "g" (cval), "g" (nval), "g" (addr));

	return ret;
}

#else
#error "ATOMIC OPERATIONS NOT DEFINED FOR TARGET ARCHITECTURE"
#endif	/* PLARCH_i386 */


#endif	/* PLATOMIC_H */
