/**
 * @file src/include/pld.h
 * PLD File structure and loader prototypes.
 */
#ifndef PLD_H
#define PLD_H

#include <plsysdep.h>
#include <pldef.h>
#include <plist.h>


/**
 * PLD Name.
 */
typedef uint32	PLDNAME;


/**
 * PLD Address.
 */
typedef union _PLDADDR {
	struct {
		PLDNAME	addr_section;
		uint32	addr_offset;
	} paddr;

	uint64 vaddr;

} __attribute__((packed)) PLDADDR;


#define PLDADDR_NULL(a)			((a)->vaddr == 0)
#define PLDADDR_WILDCARD(a)		((a)->vaddr == -1)


/**
 * PLD File.
 */
typedef struct _PLDFILE {
	PLDNAME	 pld_name;
	uint32	 pld_sections;

	/* continue sections */

} __attribute__((packed)) PLDFILE;


#define PLDFILE_NULL_ENTRY(s)		((s)->pld_name == 0 && \
					(s)->pld_sections == 0)


/**
 * PLD Section.
 */
typedef struct _PLDSECTION {
	PLDNAME	sec_name;
	uint32	sec_size;
	uint32	sec_flags;

	/* continue section contents */

} __attribute__((packed)) PLDSECTION;


#define PLDSECTION_FL_X		0x00000001
#define PLDSECTION_FL_R		0x00000002
#define PLDSECTION_FL_W		0x00000004
#define PLDSECTION_FL_Z		0x00000080

#define PLDSECTION_IS_X(s)	(((s)->sec_flags & PLDSECTION_FL_X) != 0)
#define PLDSECTION_IS_R(s)	(((s)->sec_flags & PLDSECTION_FL_R) != 0)
#define PLDSECTION_IS_W(s)	(((s)->sec_flags & PLDSECTION_FL_W) != 0)
#define PLDSECTION_IS_RW(s)	(PLDSECTION_IS_R(s) && PLDSECTION_IS_W(s))
#define PLDSECTION_IS_Z(s)	(((s)->sec_flags & PLDSECTION_FL_Z) != 0)


/**
 * EXPORT Section Entry.
 */
typedef struct _PLDEXPORT_ENTRY {
	/**
	 * Exported symbol name.
	 */
	PLDNAME	exp_symname;

	/**
	 * Address of exported symbol.
	 */
	PLDADDR	exp_symaddr;
} __attribute__((packed)) PLDEXPORT_ENTRY;


#define PLDEXPORT_NULL_ENTRY(s)		((s)->exp_symname == 0 && \
					PLDADDR_NULL(&(s)->exp_symaddr))


/**
 * IMPORT Section Entry.
 */
typedef struct _PLDIMPORT_ENTRY {
	/**
	 * Name of the imported PLD file.
	 */
	PLDNAME	imp_modname;

	/**
	 * Address containing the array of PLDIMPORT_SYMBOL.
	 * @see PLDIMPORT_SYMBOL
	 */
	PLDADDR	imp_symbols;
} __attribute__((packed)) PLDIMPORT_ENTRY;


#define PLDIMPORT_NULL_ENTRY(s)		((s)->imp_modname == 0 && \
					PLDADDR_NULL(&(s)->imp_symbols))


/**
 * IMPORTed SYMBOL.
 */
typedef struct _PLDIMPORT_SYMBOL {
	/**
	 * Name of the imported symbol.
	 */
	PLDNAME	imp_symname;

	/**
	 * Address to update with the Virtual Address of the imported symbol.
	 */
	PLDADDR imp_symaddr;
} __attribute__((packed)) PLDIMPORT_SYMBOL;


#define PLDIMPORT_NULL_SYMBOL(s)	((s)->imp_symname == 0 && \
					PLDADDR_NULL(&(s)->imp_symaddr))


/**
 * NATIVE IMPORT Section Entry.
 */
typedef struct _PLDNIMPORT_ENTRY {
	/**
	 * Address containing the asciiz string of the imported DLL.
	 */
	PLDADDR	nimp_dllname;

	/**
	 * Address containing the array of PLDNIMPORT_SYMBOLs.
	 * @see PLDNIMPORT_SYMBOL
	 */
	PLDADDR nimp_symbols;
} __attribute__((packed)) PLDNIMPORT_ENTRY;


#define PLDNIMPORT_NULL_ENTRY(s)	(PLDADDR_NULL(&(s)->nimp_dllname) && \
					PLDADDR_NULL(&(s)->nimp_symbols))


/**
 * Native IMPORTEd SYMBOL.
 */
typedef struct _PLDNIMPORT_SYMBOL {
	/**
	 * Address containing the asciiz string of the imported symbol.
	 */
	PLDADDR	nimp_symname;

	/**
	 * Address to update with the Virtual Address of the imported symbol.
	 */
	PLDADDR	nimp_symaddr;
} __attribute__((packed)) PLDNIMPORT_SYMBOL;


#define PLDNIMPORT_NULL_SYMBOL(s)	((s)->nimp_symname == 0 && \
					PLDADDR_NULL(&(s)->nimp_symaddr))


/**
 * RELOC Section Entry.
 */
typedef struct _PLDRELOC_ENTRY {
	/**
	 * Address of the referenced symbol.
	 */
	PLDADDR	rel_symbol;

	/**
	 * Address to be updated with the Virtual Addresss of the referenced
	 * symbol.
	 */
	PLDADDR	rel_address;

	/**
	 * Type of relocation to apply.
	 */
	uint32	rel_type;
} __attribute__((packed)) PLDRELOC_ENTRY;


#define PLDRELOC_NULL_ENTRY(s)		(PLDADDR_NULL(&(s)->rel_symbol) && \
					PLDADDR_NULL(&(s)->rel_address))


/**
 * Relocation Types.
 */
typedef enum _PLDRELOC_TYPE {
	/*
	 * Update by copying the absolute virtual address.
	 */
	PLDRELOC_TYPE_ABS = 0x00,

	/**
	 * Update by adding the absolute virtual address.
	 */
	PLDRELOC_TYPE_ADD = 0x01
} PLDRELOC_TYPE;


/**
 * Hash table.
 */
typedef struct _PLDHASH_ENTRY {
	/**
	 * PLD Name
	 */
	PLDNAME	hash_name;

	/**
	 * Address of the human readable (asciiz) string.
	 */
	PLDADDR	hash_string;

} __attribute__((packed)) PLDHASH_ENTRY;


#define PLDHASH_NULL_ENTRY(s)		((s)->hash_name == 0 && \
					PLDADDR_NULL(&(s)->hash_string))


/**************************************
 * Loaded (on-memory) data structures.
 **************************************/
typedef struct _PMOD	PMOD;
typedef struct _PSEC	PSEC;


/*
 * Loaded PLD Module.
 */
struct _PMOD {
	PLIST		 mod_list;	/* Should be the first member */

	PLDNAME		 mod_name;
	PLIST		 mod_seclist;	/* List of PSECs */
};


/*
 * Loaded PLD Section.
 */
struct _PSEC {
	PLIST		 se_list;	/* Should be the first memeber */

	PLDSECTION	 se_info;
	void		*se_data;	/* Contents */
};


/*
 * Prototypes.
 */
PLSTATUS	 pld_init(long *);
void		 pld_release();
PLSTATUS	 pld_open_all(const PLDFILE *);
PLSTATUS	 pld_open(void **, const PLDFILE *);
PLSTATUS	 pld_sym_any(void **, void **, PLDNAME);
PLSTATUS	 pld_sym(void **, const void *, PLDNAME);
PLSTATUS	 pld_getsec(PSEC **, const void *, PLDNAME);
void		*pld_getdata(const void *, size_t *);
void		*pld_lookup(PLDNAME);


#endif	/* PLD_H */
