/**
 * @file src/lib/lzma/lzma.c
 * Peluded lzma encoder/decoder.
 * FIXME: all the initialization parts and kernel constrains
 */
#include <plc.h>
#include <pllog.h>
#include <plzma.h>

#include "LzmaLib.h"

#define PROP_SIZE	5	/* see LzmaLib.h */


static PLSTATUS
lz2pl(int st)
{
	switch (st) {
	case SZ_OK:
		return PLSTATUS_SUCCESS;

	case SZ_ERROR_DATA:
	case SZ_ERROR_CRC:
	case SZ_ERROR_UNSUPPORTED:
	case SZ_ERROR_PARAM:
	case SZ_ERROR_INPUT_EOF:
	case SZ_ERROR_OUTPUT_EOF:
	case SZ_ERROR_READ:
		return PLSTATUS_INVALID;

	case SZ_ERROR_MEM:
		return PLSTATUS_OUT_OF_MEMORY;

	case SZ_ERROR_THREAD:
		return PLSTATUS_INTERNAL_ERROR;

	default:;
	}

	return PLSTATUS_FAILED;
}


PLSTATUS PLEXPORT
plzma_encode(long *dst, size_t *dstsiz, long src, size_t srcsiz)
{
	unsigned char *cdata, *prop;
	size_t propsiz, cdatasiz;
	long buf, err;
	PLSTATUS st;
	int lzst;

	buf = 0;
	st = plmm_realloc(&err, &buf, sizeof(uint32) + PROP_SIZE + srcsiz);
	if (st != PLSTATUS_SUCCESS) {
		return st;
	}

	prop = (unsigned char *)(buf + sizeof(uint32));
	propsiz = PROP_SIZE;

	cdata = prop + PROP_SIZE;
	cdatasiz = srcsiz;

	lzst = LzmaCompress(
	    cdata,
	    &cdatasiz,
	    (const unsigned char *)src,
	    srcsiz,
	    prop,
	    &propsiz,
	    9		/* level */,
	    1 << 26	/* dictSize */,
	    3		/* lc */,
	    0		/* lp */,
	    2		/* pb */,
	    64		/* fb */,
	    2);

	if (propsiz != PROP_SIZE) {
#ifdef PLDEBUG
		pldebug("%s: propsize=%d instead of %d\n",
		    __FUNCTION__,
		    propsiz,
		    PROP_SIZE);

		plsys_exit(-1);
		/* NOTREACHED */
#else
		plmm_free(&err, buf);
		return PLSTATUS_INTERNAL_ERROR;
#endif
	}

	if (lzst == SZ_OK) {
		*((uint32 *)buf) = PLN32H(srcsiz);
		*dstsiz = sizeof(uint32) + PROP_SIZE + cdatasiz;
		*dst = buf;
	} else {
		plmm_free(&err, buf);
	}

	return lz2pl(lzst);
}


PLSTATUS PLEXPORT
plzma_decode(long *dst, size_t *dstsiz, long src, size_t srcsiz)
{
	unsigned char *prop;
	uint32 propsiz;
	long err, buf;
	PLSTATUS st;
	int lzst;

	*dstsiz = PLN32H(*((uint32 *)src));
	
	buf = 0;
	st = plmm_realloc(&err, &buf, *dstsiz);
	if (st != PLSTATUS_SUCCESS) {
		return st;
	}

	prop = (unsigned char *)src + sizeof(uint32);
	propsiz = PROP_SIZE;

	src += sizeof(uint32) + PROP_SIZE;
	srcsiz -= (sizeof(uint32) + PROP_SIZE);

	lzst = LzmaUncompress(
	    (unsigned char *)buf,
	    dstsiz,
	    (const unsigned char *)src,
	    &srcsiz,
	    (const unsigned char *)prop,
	    5);

	if (lzst == SZ_OK) {
		*dst = buf;
	} else {
		plmm_free(&err, buf);
	}

	return lz2pl(st);
}


#define kNumStates 12
extern Byte kLiteralNextStates[kNumStates * 2];
extern void *SzAlloc(void *, size_t);
extern void SzFree(void *, void *);
extern ISzAlloc g_Alloc;


PLSTATUS PLEXPORT
plzma_init()
{
#ifndef PLSTATIC
	kLiteralNextStates[ 0] = 0;
	kLiteralNextStates[ 1] = 0;
	kLiteralNextStates[ 2] = 0;
	kLiteralNextStates[ 3] = 0;
	kLiteralNextStates[ 4] = 1;
	kLiteralNextStates[ 5] = 2;
	kLiteralNextStates[ 6] = 3;
	kLiteralNextStates[ 7] = 4;
	kLiteralNextStates[ 8] = 5;
	kLiteralNextStates[ 9] = 6;
	kLiteralNextStates[10] = 4;
	kLiteralNextStates[11] = 5;
	kLiteralNextStates[12] = 7;
	kLiteralNextStates[13] = 7;
	kLiteralNextStates[14] = 7;
	kLiteralNextStates[15] = 7;
	kLiteralNextStates[16] = 7;
	kLiteralNextStates[17] = 7;
	kLiteralNextStates[18] = 7;
	kLiteralNextStates[19] = 10;
	kLiteralNextStates[20] = 10;
	kLiteralNextStates[21] = 10;
	kLiteralNextStates[22] = 10;
	kLiteralNextStates[23] = 10;

	g_Alloc.Alloc = SzAlloc;
	g_Alloc.Free = SzFree;
#endif
	return PLSTATUS_SUCCESS;
}

 
