/**
 * @file src/lib/plc/plinit.c
 */
#include <plc.h>
#include <plsock.h>


extern PLSTATUS plstd_init();

int  PLEXPORT *(*plgeterrno)() = NULL;
int  PLEXPORT *(*plgetherrno)() = NULL;
void PLEXPORT (*plenter_cancel_point)() = NULL;
void PLEXPORT (*plleave_cancel_point)() = NULL;
void PLEXPORT (*plprepare_fork)() = NULL;
void PLEXPORT (*plparent_fork)() = NULL;
void PLEXPORT (*plchild_fork)() = NULL;
				/*
				 * Can't have a default value other than NULL:
				 * See kernel coding constraints documentation.
				 */

/**
 * Environment (see crt0.c).
 */
char PLEXPORT **environ = NULL;


/**
 * Errno variable used on unthreaded processes.
 */
static int _errno = 0;


/**
 * h_errno variable used on unthreaded processes.
 */
static int _h_errno = 0;


/**
 * Return the address of the unthreaded errno variable.
 *
 * @see _errno
 */
static int *
pl_unthreaded_errno()
{
	return &_errno;
}


/**
 * Return the address of the unthreaded h_errno variable.
 *
 * @see _h_errno
 */
static int *
pl_unthreaded_h_errno()
{
	return &_h_errno;
}


/**
 * Dummy function called when entering or
 * leaving cancellation point functions
 * on single-thread applications.
 */
static void
pl_unthreaded_cancel()
{
}


/**
 * Set cancellation point callbacks.
 * These callbacks are called by cancellation points functions defined
 * by POSIX.
 * <br>
 * This function is intended to be called only by a multi-threading library
 * during its initialization.
 * <br>
 * Follows the list of cancellation point functions defined by the standard:
 *	- accept()
 *	- aio_suspend()		not supported
 *	- clock_nanosleep()	not supported
 *	- close()
 *	- connect()
 *	- creat()		not supported
 *	- fcntl()
 *	- fdatasync()
 *	- fsync()
 *	- getmsg()		not supported
 *	- getpmsg()		not supported
 *	- lockf()		not supported
 *	- mq_receive()		not supported
 *	- mq_send()		not supported
 *	- mq_timedreceive()	not supported
 *	- mq_timedsend()	not supported
 *	- msgrcv()		not supported
 *	- msgsnd()		not supported
 *	- msync()
 *	- nanosleep()		XXX
 *	- open()
 *	- pause()		not supported
 *	- poll()
 *	- pread()		not supported
 *	- pselect()		not supported
 *	- pthread_cond_timedwait()
 *	- pthread_cond_wait()
 *	- pthread_join()	XXX
 *	- pthread_testcancel()
 *	- putmsg()		not supported
 *	- putpmsg()		not supported
 *	- pwrite()		not supported
 *	- read()
 *	- readv()
 *	- recv()
 *	- recvfrom()
 *	- recvmsg()
 *	- select()
 *	- sem_timedwait()	not supported
 *	- sem_wait()
 *	- send()
 *	- sendmsg()
 *	- sendto()
 *	- sigpause()		not supported
 *	- sigsuspend()
 *	- sigtimedwait()	not supported
 *	- sigwait()		XXX
 *	- sigwaitinfo()		not supported
 *	- sleep()		not supported
 *	- system()		not supported
 *	- tcdrain()		not supported
 *	- usleep()		not supported
 *	- wait()		not supported
 *	- waitid()		not supported
 *	- waitpid()		not supported
 *	- write()
 *	- writev()
 *
 * @param enter (IN) If not <code>NULL</code>, the new enter callback;
 *
 * @param leave (IN) If not <code>NULL</code>, the new leave callback;
 *
 * @param oenter (OUT) If not <code>NULL</code> updated
 * with the old enter callback;
 *
 * @param oleave (OUT) If not <code>NULL</code> updated
 * with the old leave callback.
 *
 * @see http://www.opengroup.org/onlinepubs/009695399/functions/xsh_chap02_09.html
 * @see plenter_cancel_point()
 * @see plleave_cancel_point()
 */
void PLEXPORT
plsetcancel(void (*enter)(), void (*leave)(), void (**oenter)(),
    void (**oleave)())
{
	if (oenter != NULL) {
		*oenter = plenter_cancel_point;
	}

	if (enter != NULL) {
		plenter_cancel_point = enter;
	}

	if (oleave != NULL) {
		*oleave = plleave_cancel_point;
	}

	if (leave != NULL) {
		plleave_cancel_point = leave;
	}
}


PLSTATUS PLEXPORT
plinit()
{
	PLSTATUS st;

	if ((st = plsys_init()) == PLSTATUS_SUCCESS) {

		plseterrno(pl_unthreaded_errno);
		plsetherrno(pl_unthreaded_h_errno);
		plsetcancel(pl_unthreaded_cancel, pl_unthreaded_cancel,
		    NULL, NULL);

		/*
		 * Calling order matters.
		 */
		if ((st = plmm_init()) == PLSTATUS_SUCCESS &&
		    (st = plsock_init()) == PLSTATUS_SUCCESS &&
		    (st = plstd_init()) == PLSTATUS_SUCCESS);
	}

	return st;
}


