/**
 * @file src/lib/plc/plvnprint_cb.c
 */
#include <plc.h>


/**
 * Get a string containing the hexadecimal representation of an
 * unisigned 32 bits value.
 *
 * @param a (IN) Unsigned 32 bits value;
 *
 * @return A static buffer containing the hexadecimal string is returned.
 */
static char *
_u32toh(uint32 a)
{
	static char buf[] = { "0123456789ABCDEF" };
	static char dst[10] = { 0 };

	*(dst + 7) = *(buf + (a         & 0xF));
	*(dst + 6) = *(buf + ((a >>  4) & 0xF));
	*(dst + 5) = *(buf + ((a >>  8) & 0xF));
	*(dst + 4) = *(buf + ((a >> 12) & 0xF));
	*(dst + 3) = *(buf + ((a >> 16) & 0xF));
	*(dst + 2) = *(buf + ((a >> 20) & 0xF));
	*(dst + 1) = *(buf + ((a >> 24) & 0xF));
	*(dst + 0) = *(buf + ((a >> 28) & 0xF));
	*(dst + 8) = '\0';

	return dst;
}


/**
 * Get a string containing the decimal representation of an
 * unsigned 32 bits value.
 *
 * @param a (IN) Unsigned 32 bits value;
 *
 * @return A static buffer containing the decimal string is returned.
 */
static char *
_u32toa(uint32 a)
{
	static char dst[20] = { 0 };
	char *p1, *p2, c;
	int i, s;

	i = 0;
	s = (int)a;

	if (s < 0) {
		a = -a;
	}

	do {
		*(dst + i++) = '0' + a % 10;
		a /= 10;
	} while (a);

	if (s < 0) {
		*(dst + i++) = '-';
	}

	*(dst + i) = '\0';

	for (p1 = dst, p2 = dst + i - 1; p1 < p2; p1++, p2--) {
		c = *p1;
		*p1 = *p2;
		*p2 = c;
	}

	return dst;
}


/**
 * Print a formatted string using a given callback.
 * Recognized format parameters:
 * <br>
 * <b>%s</b>: The argument is a string, just print it as is;<br>
 *
 * <b>%d</b>, <b>%u</b>: The argument is an unsigned 32 bits value,
 * print its decimal representation;<br>
 *
 * <b>%x</b>, <b>%p</b>: The argument is an unsigned 32 bits value, print its
 * hexadecimal representation adding dealing zeores;<br>
 *
 * <b>%X</b>: The argument is an unsigned 64 bits value, print its
 * hexadecimal representation adding dealing zeroes.
 *
 * @param fmt (IN) Contains a string specifing how to format the
 * output string;
 *
 * @param args (IN) List of arguments;
 *
 * @param doprint (IN) Callback printing function;
 *
 * @param par (IN) Argument to send to the callback printing function;
 *
 * @param siz (IN) Size
 *
 * @see plva_list
 * @see _u32toa()
 * @see _u32toh()
 *
 * @todo Document better
 */
int
plvnprint_cb(const char *fmt, plva_list args,
    int (*doprint)(void *, const char *, size_t), void *par, size_t siz)
{
	char tmp[2] = { 0, 0 };
	int count, printed;
	uint32 n;

	count = 0;

	for (; siz > 0 && *fmt != '\0'; fmt++) {
		printed = 0;
		if (*fmt == '%') {
			fmt++;

			switch (*fmt) {
			case '\0':
				return 0;

			case 'c':
				tmp[0] = plva_arg(args, uint32);
				printed = doprint(par, tmp, siz);
				break;

			case 'd':
			case 'u':
				printed = doprint(par,
				    _u32toa(plva_arg(args, uint32)), siz);
				break;

			case 's':
				printed += doprint(par,
				    (char *)plva_arg(args, char *), siz);
				break;

			case 'x':
			case 'p':
				printed += doprint(par,
				    _u32toh(plva_arg(args, uint32)), siz);
				break;

			case 'X':
				n = plva_arg(args, uint32);
				printed += doprint(par,
				    _u32toh(plva_arg(args, uint32)), siz);
				printed += doprint(par, _u32toh(n), siz);
				break;

			default:
				goto copy;
			}

		} else {
copy:			tmp[0] = *fmt;
			count += doprint(par, tmp, siz);
		}

		siz -= printed;
		count += printed;
	}

	return count;
}


