/**
 * @file src/lib/plc/std/signal_sigwait.c
 */
#include <plc.h>
#include <pllog.h>

#include <signal.h>


/**
 * Suspend the calling thread until a signal is delivered to it.
 *
 * @param set (IN) The calling thread is suspended until one of the
 * signals specified in this set arrives;
 *
 * @param signo (OUT) If not <code>NULL</code> it is updated with the
 * signal that wokeup the thread.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * NOTE from linux manpage:
 * For sigwait to work reliably, the signals being waited for must be
 * blocked in all threads, not only in the calling thread, since otherwise
 * the POSIX semantics for signal delivery do not guarantee that it’s the
 * thread doing the sigwait that will receive the signal. The best way
 * to achieve this is block those signals before any threads are created,
 * and never unblock them in the program other than by calling sigwait.
 *
 * @see plt_kill()
 * @see pthread_sigmask()
 * @see plsys_tsigwait()
 */
int PLEXPORT
sigwait(const sigset_t *set, int *signo)
{
	PLSTATUS st;

	plenter_cancel_point();

	while ((st = plsys_tsigwait(set, signo)) != PLSTATUS_SUCCESS) {
		if (st == PLSTATUS_TRY_AGAIN ||
		    (st == PLSTATUS_FAILED && errno == EINTR)) {
/* Too noisy */
#if 0
			pldebug("%s: Interrupted. Trying again...\n",
			    __FUNCTION__);
#endif
			continue;
		}

		pldebug("%s: Failed, peludo status=<%s>, errno=%d\n",
		    __FUNCTION__, plstrstatus(st), errno);

		break;
	}

	plleave_cancel_point();

	return ((st == PLSTATUS_SUCCESS) ? 0 : errno);
}


