/**
 * @file src/lib/plfs/plfs.c
 * Peludo Simple Memory File System.
 * The author don't like this code.
 *
 * TODO: Review the locking mechanism.
 * FIXME: Lock of handle by the fs manager (PLVFS)
 * FIXME: Checking of invalid handles, entries, etc. \
 * 	  splitted between fs manager and plfs.
 * FIXME: Do it better
 */
#include <plc.h>
#include <plspinlock.h>
#include <plist.h>
#include <plfs.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>


/**
 * A file node into the list.
 */
typedef struct _PLFS_ENTRY PLFS_ENTRY;
struct _PLFS_ENTRY {
	/**
	 * Prev/Next entries belonging to the same parent.
	 */
	PLIST list;		/* Should be the first member */

	/**
	 * Entry lock.
	 */
	PLOCK lock;		/* XXX Should be a RW lock */

	/**
	 * Name of the entry.
	 */
	char name[PLFS_NAME_MAX];

	/**
	 * Parent entry.
	 * The parent is always a directory entry.
	 */
	PLFS_ENTRY *parent;

	/**
 	 * Directory or File flag.
 	 */
	int isdir;

	/**
	 * Entry contents (depends on <code>isdir</code>).
	 */
	union {
		PLIST childs;	/* List of child entries */
		void *data;	/* Pointer to the file data */
	} contents;

	/**
	 * Available data bytes or number of childs.
	 */
	uint32 size;

	/*
	 * Number of used bytes (valid for file entries).
	 */
	uint32 used;

	/**
	 * Number of open handles to this entry.
	 */
	int opencount;

	/**
	 * Temporal file
	 * A non-zero value indicates that entry should be removed
	 * when <code>opencount</code> reaches 0 and, if it is a directory,
	 * when it becomes empty. 
	 */
	int temporal;

};


/**
 * Handle to an open file.
 *
 * @see PLFS_FILE
 */
typedef struct _PLFS_HANDLE {
	PLOCK		 lock;		/* To make it thread safe */
	PLFS_ENTRY	*entry;		/* The open entry */
	unsigned long	 offset;	/* File offset/Dir trav counter */
	uint32		 desize;	/* Size of de array */
	struct dirent	*de;		/* Directory entry array */
} PLFS_HANDLE;


/**
 * The root entry.
 * @see PLFS_ENTRY
 */
/*
 * This is not a real initialization,
 * just set members to 0 to avoid .bss
 * (see kernel coding constraints).
 */
static PLFS_ENTRY plfs_root = {
	.list = PLIST_INITIALIZED,
	.lock = PLOCK_INITIALIZED,
	.name = { 0 },
	.parent = NULL,
	.isdir = 0,
	.contents.childs = PLIST_INITIALIZED,
	.size = 0,
	.used = 0,
	.opencount = 0,
	.temporal = 0
};


/**
 * Search for a child entry.
 *
 * @param name (IN) Name of the child (not delimited by any
 * <code>PLFS_DIR_SEPARATOR</code> character);
 *
 * @param parent (IN) Parent directory (<b>should be locked in advance</b>).
 *
 * @return The child entry or <code>NULL</code> if it does not exist.
 *
 * @see PLFS_DIR_SEPARATOR
 * @see PLFS_ENTRY
 */
static PLFS_ENTRY *
_plfs_lookup_child(const char *name, const PLFS_ENTRY *parent)
{
	PLFS_ENTRY *entry;

	PLIST_FOREACH(&parent->contents.childs, entry, list) {
		if (!strcmp(entry->name, name)) {
			return entry;
		}
	}

	return NULL;
}


/**
 * Search for an entry.
 *
 * @param path (IN/OUT) Fullpath of the desired entry
 * (should not be and should not point to <code>NULL</code>;
 * the contents of this buffer could change at exit);
 *
 * @param parent (IN) Parent directory (should not be <code>NULL</code>).
 *
 * @return If the entry exists it is locked and a pointer to it is returned;
 * otherwise <code>errno</code> is set and <code>NULL</code> is returned.
 *
 * @see PLFS_ENTRY
 * @see plfs_root
 */
static PLFS_ENTRY *
_plfs_lookup(char **path, PLFS_ENTRY *parent)
{
	PLFS_ENTRY *entry;
	char *name;

	for (; **path != '\0' && **path == PLFS_DIR_SEPARATOR_CHAR; (*path)++);

	name = strsep(path, PLFS_DIR_SEPARATOR);

	plock_splock(&parent->lock);

	entry = _plfs_lookup_child(name, parent);
	if (entry != NULL) {
		if (*path == NULL || **path == '\0') {
			plock_splock(&entry->lock);
		} else {
			entry = (entry->isdir) ?
			    _plfs_lookup(path, entry) : NULL;
		}
	}

	plock_unlock(&parent->lock);		

	return entry;
}


/**
 * Search for a parent directory.
 *
 * @param path (IN) Fullpath of the entry whose parent is to be found
 * (should not be and should not point to <code>NULL</code>;
 * the contents of this buffer could change at exit);
 *
 * @param name (OUT) Updated with the basename inside <code>path</code>.
 *
 * @return If the parent directory exists it is locked and a pointer
 * to it is returned; otherwise <code>errno</code> is set and
 * <code>NULL</code> is returned.
 *
 * @see PLFS_ENTRY
 */
static PLFS_ENTRY *
_plfs_lookup_parent(char **path, char **basename)
{
	char *p;

	/*
	 * Get the parent directory string.
	 */
	for (p = *path + strlen(*path) - 1;
	    p != *path && *p != PLFS_DIR_SEPARATOR_CHAR && *p != '\0';
	    p--);

	if (p == *path) {
		/*
		 * The parent is the root entry.
		 */
		plock_splock(&plfs_root.lock);
		*basename = p + ((*p == PLFS_DIR_SEPARATOR_CHAR) ? 1 : 0);
		return &plfs_root;
	}

	*p = '\0';
	*basename = p + 1;

	return _plfs_lookup(path, &plfs_root);
}


/**
 * Create a new entry and add it inside the parent child list.
 * <b>The parent entry should be locked in advance</b>.
 * It is not checked if the parent is a valid directory entry.
 * The new entry will be created as an empty file.
 *
 * @param name (IN) Name of the entry (not delimited by any
 * <code>PLFS_DIR_SEPARATOR</code> character);
 *
 * @param parent (IN) Parent entry (should not be <code>NULL</code>).
 *
 * @return On success the new entry is locked and a pointer to it returned;
 * otherwise <code>errno</code> is set and <code>NULL</code> is returned.
 *
 * @see PLFS_DIR_SEPARATOR
 * @see PLFS_ENTRY
 */
static PLFS_ENTRY *
_plfs_create_child(const char *name, PLFS_ENTRY *parent)
{
	PLFS_ENTRY *entry;

	/*
	 * Create the new file entry.
	 */
	entry = (PLFS_ENTRY *)malloc(sizeof(*entry));
	if (entry != NULL) {
		bzero(entry, sizeof(*entry));
		PLIST_INIT(&entry->list);
		entry->lock = PLOCK_INITIALIZED;
		strncpy(entry->name, name, sizeof(entry->name));
		entry->parent = parent;

		plock_splock(&entry->lock);

		/*
		 * Add the new entry inside the parent's child list.
		 */
		plist_add_tail((PLIST *)&parent->contents.childs, (PLIST *)entry);
		parent->size++;
	}

	return entry;
}


/**
 * Create an entry into the PLFS.
 *
 * @param name (IN) Name of the entry (not delimited by any
 * <code>PLFS_DIR_SEPARATOR</code> character);
 *
 * @param parent (IN) Parent entry (should not be <code>NULL</code>);
 *
 * @param isdir (IN) A non-zero value indicates that the entry should
 * be a directory;
 *
 * @param temp (IN) A non-zero value indicates that the entry should be
 * created as temporal (be removed when the last open handle is closed).
 *
 * @return On success the new entry is locked and a pointer to it returned;
 * otherwise <code>errno</code> is set and <code>NULL</code> is returned.
 *
 * @see PLFS_DIR_SEPARATOR
 * @see PLFS_ENTRY
 * @see _plfs_create_child()
 */
static PLFS_ENTRY *
_plfs_create_entry(const char *name, PLFS_ENTRY *parent, int isdir, int temp)
{
	PLFS_ENTRY *entry;

	entry = _plfs_create_child(name, parent);
	if (entry != NULL) {
		entry->temporal = temp;
		if (isdir) {
			entry->isdir = 1;
			PLIST_INIT(&entry->contents.childs);
		}
	}

	return entry;
}


/**
 * Resize a data file entry.
 * @param entry (IN) Entry to resize (should be a <b>file</b> entry
 * <b>locked</b> in advance);
 *
 * @param size (IN) New size (in bytes).
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see PLFS_ENTRY
 */
static PLSTATUS
_plfs_resize_file(PLFS_ENTRY *entry, uint32 size)
{
	void *data;

	/*
	 * Release all data if the desired size is 0.
	 */
	if (size == 0) {
		if (entry->size) {
			free(entry->contents.data);
			entry->contents.data = NULL;
			entry->size = 0;
			entry->used = 0;
		}

		return PLSTATUS_SUCCESS;
	}

	/*
	 * Alloc requested number of blocks. 
	 */
	size = PLALIGN(size, PLFS_BLOCK_SIZE);
	if (size != entry->size) {
		data = realloc(entry->contents.data, size);
		if (data == NULL) {
			return PLSTATUS_FAILED;
		}

		if (size > entry->size) {
			bzero((uint8 *)data + entry->used, size - entry->used);
		}

		entry->contents.data = data;
		entry->size = size;
	}

	/*
	 * Truncate if shrink.
	 */
	if (entry->used > size) {
		entry->used = size;
	}

	return PLSTATUS_SUCCESS;
}


/**
 * Remove an entry from the PLFS.
 * The entry to delete and its parent should both <b>be locked in advance</b>.
 *
 * @param entry (IN) Entry to delete
 * (if it is a directory it should not have childs).
 *
 * @return A Peludo status code.
 *
 * @see PLFS_ENTRY
 */
static void
_plfs_delete_entry(PLFS_ENTRY *entry)
{
	/*
	 * Remove the entry from the parent's child list.
	 */
	plist_del(&entry->list);
	entry->parent->size--;

	/*
	 * Release the entry.
	 */
	if (!entry->isdir && entry->contents.data != NULL) {
		free(entry->contents.data);
	}

	free(entry);
}


/**
 * Open or create a file or directory.
 *
 * @param handle (OUT) Updated with a handle to the open file
 * (should not be <code>NULL</code>);
 *
 * @param path (IN) Full path name;
 *
 * @param flags (IN) Open flags (OR combination of:
 * <code>O_CREAT</code>,
 * <code>O_EXCL</code>,
 * <code>O_DIRECTORY</code>,
 * <code>O_TRUNC</code>,
 * <code>O_APPEND</code>);
 *
 * @param mode (IN) Creation mode (ignored).
 *
 * @param temporal (IN) A non-zero value indicates that a newly
 * created entry should be removed once all handles associated to it
 * are closed; if the entry is a directory it will be removed when
 * it becomes empty.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 */
static PLSTATUS
plfs_open(long *handle, const char *path, long flags, long mode, int temporal)
{
	char fullpath[PATH_MAX], *p, *basename;
	PLFS_ENTRY *parent, *entry;
	PLFS_HANDLE *h;
	int created;

	/*
	 * Create the handle.
	 */
	h = (PLFS_HANDLE *)malloc(sizeof(PLFS_HANDLE));
	if (h == NULL) {
		return PLSTATUS_FAILED;
	}

	/*
	 * Copy the path into writable memory.
	 */
	strncpy(fullpath, path, sizeof(fullpath));
	fullpath[sizeof(fullpath) - 1] = '\0';
	p = fullpath;

	/*
 	 * Get exising parent and base entries.
 	 */
	parent = _plfs_lookup_parent(&p, &basename);
	if (parent == NULL) {
		return PLSTATUS_FAILED;
	}

	/*
	 * Get the entry to open.
	 */
	if (*basename == '\0') {
		/*
		 * It is trying to open the root directory:
		 * It cannot be created.
		 */
		entry = parent;
		flags &= ~O_CREAT;
	} else {
		entry = _plfs_lookup_child(basename, parent);
	}

	if (entry != NULL) {
		/*
		 * We are sure that the entry pointer is valid
		 * because its parent is locked.
		 */
		if (entry != parent) {
			plock_splock(&entry->lock);
		}

		created = 0;

		/*
		 * The entry exists and it should not?
		 */
		if ((flags & (O_CREAT | O_EXCL)) == (O_CREAT | O_EXCL)) {
			errno = EEXIST;
			goto faile;
		}

		/*
		 * Should it be a directory?
		 */
		if ((flags & O_DIRECTORY) && !entry->isdir) {
			errno = ENOTDIR;
			goto faile;
		}

	} else {
		/*
		 * The entry does not exist: Create it?
		 */
		if (!(flags & O_CREAT)) {
			errno = ENOENT;
			goto fail;
		}

		entry = _plfs_create_entry(basename, parent, flags & O_DIRECTORY, temporal);
		if (entry == NULL) {
			goto fail;
		}

		created = 1;
	}

	/*
	 * Truncate.
	 */
	if ((flags & O_TRUNC) && entry->size != 0) {
		if (entry->isdir) {
			/*
			 * A directory cannot be truncated.
			 */
			errno = ENOTSUP;
			goto faile;
		}

		if (_plfs_resize_file(entry, 0) != PLSTATUS_SUCCESS) {
			goto faile;
		}
	}

	/*
	 * Open counter.
	 */
	entry->opencount++;

	plock_unlock(&entry->lock);

	if (entry != parent) {
		plock_unlock(&parent->lock);
	}

	/*
	 * Update handle members.
	 */
	PLOCK_INIT_UNLOCKED(&h->lock);
	h->entry = entry;
	h->offset = (!entry->isdir && (flags & O_APPEND)) ? entry->used : 0;
	h->desize = 0;
	h->de = NULL;
	*handle = (long)h;

	return PLSTATUS_SUCCESS;

faile:	if (created) {
		_plfs_delete_entry(entry);
	} else {
		plock_unlock(&entry->lock);
	}

fail:	if (entry != parent) {
		plock_unlock(&parent->lock);
	}

	free((void *)h);

	return PLSTATUS_FAILED;
}


/**
 * Close an open file or directory.
 * The given handle is to be destroyed so the caller
 * must ensure that no other thread is using it.
 * <br>
 * If the entry is a temporal file (see plfs_open()) 
 * and there are not other open handles to it, it is
 * deleted from the PLFS.
 *
 * @param handle (IN) Valid handle to the open entry.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see plfs_open()
 */
static PLSTATUS
plfs_close(long handle)
{
	PLFS_ENTRY *parent;
	PLFS_HANDLE *h;

	h = (PLFS_HANDLE *)handle;

	plock_splock(&h->lock);
	plock_splock(&h->entry->lock);

	if (h->entry->opencount == 1 && h->entry->temporal) {
		/*
		 * Cannot remove non-empty directories.
		 */
		if (!h->entry->isdir ||
		    (h->entry->isdir && h->entry->size == 0)) {
			/*
			 * Remove the unreferenced temporal entry.
			 *
			 * To remove the entry the parent must be locked,
			 * but this lock should be done before the entry.
			 *
			 * The following code is safe because we still own
			 * an open handle to the file.
			 */
			parent = h->entry->parent;
			plock_unlock(&h->entry->lock);

			plock_splock(&parent->lock);
			plock_splock(&h->entry->lock);

			_plfs_delete_entry(h->entry);

			plock_unlock(&parent->lock);

			goto bye;
		}
	}

	/*
	 * Decrease the opencounter and destroy the handle.
	 */
	h->entry->opencount--;
	plock_unlock(&h->entry->lock);

bye:	free(h);

	return PLSTATUS_SUCCESS;
}


/**
 * Change the file offset position of an open file.
 * 
 * @param handle (IN) Valid handle to the open entry;
 *
 * @param offset (IN) Offset to be set;
 *
 * @param newoff (OUT) If not <code>NULL</code> it is updated with
 * the resulting offset (from the start of the file).
 *
 * @param whence (IN) indicates how to treat the specified offset;
 * it should be one One of: <code>SEEK_SET</code>, <code>SEEK_CUR</code>,
 * <code>SEEK_END</code>.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 */
static PLSTATUS
plfs_lseek(long handle, off_t offset, int whence, off_t *newoff)
{
	PLFS_HANDLE *h;
	int64 noff;

	h = (PLFS_HANDLE *)handle;
	plock_splock(&h->lock);

	switch (whence) {
	case SEEK_SET:
		noff = offset;
		break;

	case SEEK_CUR:
		noff = offset + h->offset;
		break;

	case SEEK_END:
		plock_splock(&h->entry->lock);
		noff = offset + h->entry->used;
		plock_unlock(&h->entry->lock);
		break;

	default:
		noff = -1;
	}

	if (noff < 0 || noff > PLFS_MAXSIZE) {
		errno = EINVAL;
		plock_unlock(&h->lock);
		return PLSTATUS_FAILED;
	}

 	/*
	 * http://www.opengroup.org/onlinepubs/000095399/functions/lseek.html
	 * [...] The lseek() function shall allow the file offset to be set
	 * beyond the end of the existing data in the file. If data is later
	 * written at this point, subsequent reads of data in the gap shall
	 * return bytes with the value 0 until data is actually written into
	 * the gap.
	 *
	 * See plfs_write().
	 */
	h->offset = (uint32)noff;
	plock_unlock(&h->lock);

	if (newoff != NULL) {
		*newoff = (off_t)noff;
	}

	return PLSTATUS_SUCCESS;
}


/**
 * Remove a file or an empty directory.
 * If there are other open handles to the same entry
 * it will not be removed until the last handle is closed
 * (if it is a directory it will be removed when get empty).
 *
 * @param path (IN) Fullpath of the entry to remove.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see plfs_open()
 * @see plfs_close()
 */
static PLSTATUS
plfs_unlink(const char *path)
{
	PLFS_HANDLE *h;
	long handle;
	PLSTATUS st;

	st = plfs_open(&handle, path, 0, 0, 0);

	if (st == PLSTATUS_FAILED && errno == ENOENT) {
		return PLSTATUS_SUCCESS;
	}

	if (st == PLSTATUS_SUCCESS) {
		/*
		 * Force deletion by setting it as temporal.
		 */
		h = (PLFS_HANDLE *)handle;
		plock_splock(&h->entry->lock);
		h->entry->temporal = 1;
		plock_unlock(&h->entry->lock);

		st = plfs_close(handle);
	}

	return st;
}


/**
 * Read from a file handle.
 *
 * @param handle (IN) Handle to the open file
 * (it should be associated to a file entry);
 *
 * @param buf (OUT) Buffer to fill with the read data;
 *
 * @param bufsiz (IN) Maximum number of bytes to read;
 *
 * @param read (OUT) If not <code>NULL</code>,
 * updated with the number of read bytes.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see plfs_lseek()
 * @see plfs_open()
 * @see plfs_write()
 */
static PLSTATUS
plfs_read(long handle, void *buf, size_t bufsiz, size_t *read)
{
	PLFS_HANDLE *h;
	uint32 tocpy;

	if (bufsiz == 0) {
		*read = 0;
		return PLSTATUS_SUCCESS;
	}

	h = (PLFS_HANDLE *)handle;

	plock_splock(&h->lock);
	plock_splock(&h->entry->lock);

	/*
	 * Cannot read from a directory entry.
	 */
	if (h->entry->isdir) {
		errno = EISDIR;
		plock_unlock(&h->entry->lock);
		plock_unlock(&h->lock);
		return PLSTATUS_FAILED;
	}

	/*
	 * All data already read?
	 */
	if (h->offset < h->entry->used) {
		tocpy = h->entry->used - h->offset;
		if (bufsiz < tocpy) {
			tocpy = bufsiz;
		}
		memcpy(buf, (uint8 *)h->entry->contents.data + h->offset, tocpy);
		*read = (size_t)tocpy;
		h->offset += tocpy;
	} else {
		if (read != NULL) {
			*read = 0;
		}
	}
			
	plock_unlock(&h->entry->lock);
	plock_unlock(&h->lock);

	return PLSTATUS_SUCCESS;
}


/**
 * Write data into a file.
 *
 * @param handle (IN) Handle to the open file
 * (it should be associated to a file entry);
 *
 * @param buf (OUT) Buffer to write;
 *
 * @param bufsiz (IN) Number of bytes to write;
 *
 * @param written (OUT) If not <code>NULL</code>,
 * updated with the number of written bytes.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see plfs_lseek()
 * @see plfs_open()
 * @see plfs_write()
 */
static PLSTATUS
plfs_write(long handle, const void *buf, size_t bufsiz, size_t *written)
{
	PLFS_HANDLE *h;
	uint64 size;
	PLSTATUS st;

	if (bufsiz == 0) {
		*written = 0;
		return PLSTATUS_SUCCESS;
	}

	h = (PLFS_HANDLE *)handle;
	st = PLSTATUS_FAILED;

	plock_splock(&h->lock);

	/*
	 * Check maximum file size.
	 */
	size = (uint64)h->offset + (uint64)bufsiz;
	if (size > PLFS_MAXSIZE) {
		errno = ENOSPC;
		goto bye2;
	}

	plock_splock(&h->entry->lock);

	/*
	 * Cannot write to a directory entry.
	 */
	if (h->entry->isdir) {
		errno = EISDIR;
		goto bye;
	}

	/*
	 * If needed expand the entry size.
	 */
	if (size > h->entry->size) {
		st = _plfs_resize_file(h->entry, (uint32)size);
		if (st != PLSTATUS_SUCCESS) {
			goto bye;
		}
	}

	/*
	 * Write data.
	 */
	memcpy((uint8 *)h->entry->contents.data + h->offset, buf, bufsiz);
	h->entry->used = h->offset + (uint32)bufsiz;
	h->offset += (uint32)bufsiz;

	if (written != NULL) {
		*written = (uint32)bufsiz;
	}

	st = PLSTATUS_SUCCESS;

bye:	plock_unlock(&h->entry->lock);
bye2:	plock_unlock(&h->lock);

	return st;
}


/**
 * Retrieve information about an open entry.
 *
 * @param handle (IN) Handle to the open entry;
 *
 * @param buf (OUT) Updated with the entry state.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see stat
 */
static PLSTATUS
plfs_fstat(long handle, struct stat *buf)
{
	PLFS_HANDLE *h;

	h = (PLFS_HANDLE *)handle;

	bzero(buf, sizeof(struct stat));

	plock_splock(&h->lock);
	plock_splock(&h->entry->lock);

	bzero(buf, sizeof(*buf));
	if (h->entry->isdir) {
		buf->st_size = h->entry->size;
		buf->st_mode = S_IFDIR;
		buf->st_blksize = sizeof(PLFS_ENTRY);
		buf->st_blocks = h->entry->size;
	} else {
		buf->st_size = h->entry->used;
		buf->st_mode = S_IFREG;
		buf->st_blksize = PLFS_BLOCK_SIZE;
		buf->st_blocks = h->entry->size / PLFS_BLOCK_SIZE;
	}

	plock_unlock(&h->entry->lock);
	plock_unlock(&h->lock);

	return PLSTATUS_SUCCESS;
}


/**
 * Retrieve information about an entry.
 *
 * @param path (IN) Fullpath to the entry;
 *
 * @param buf (OUT) Updated with the entry state.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see plfs_fstat()
 */
static PLSTATUS
plfs_stat(const char *path, struct stat *buf)
{
	long handle;
	PLSTATUS st;

	st = plfs_open(&handle, path, 0, 0, 0);
	if (st == PLSTATUS_SUCCESS) {
		st = plfs_fstat(handle, buf);
		(void)plfs_close(handle);
	}

	return st;
}


/**
 * XXX
 */
static PLSTATUS
plfs_ftruncate(long handle, off_t size)
{
	PLFS_HANDLE *h;
	PLSTATUS st;

	if (size > PLFS_MAXSIZE) {
		errno = ENOSPC;
		return PLSTATUS_FAILED;
	}

	h = (PLFS_HANDLE *)handle;

	plock_splock(&h->lock);
	plock_splock(&h->entry->lock);

	if (h->entry->isdir) {
		errno = EISDIR;
		st = PLSTATUS_FAILED;
	} else {
		st = _plfs_resize_file(h->entry, size);
	}

	plock_unlock(&h->entry->lock);
	plock_unlock(h->lock);

	return st;
}


/**
 * Rename an entry.
 */
static PLSTATUS
plfs_rename(const char *old, const char *new)
{
	errno = ENOSYS;
	return PLSTATUS_NOT_IMPLEMENTED;
}


/**
 * Open a directory entry.
 *
 * @param handle (OUT) Updated with a handle to the open entry;
 *
 * @param path (IN) Path to the directory to open.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see plfs_open()
 */
static PLSTATUS
plfs_opendir(long *handle, const char *path)
{
	struct dirent *de;
	PLFS_HANDLE *h;
	PLFS_ENTRY *e;
	PLSTATUS st;
	long hnd;

	/*
	 * Open the directory entry.
	 */
	st = plfs_open(&hnd, path, O_DIRECTORY, 0, 0);
	if (st != PLSTATUS_SUCCESS) {
		return st;
	}

	h = (PLFS_HANDLE *)hnd;

	plock_splock(&h->entry->lock);

	/*
	 * Fill the handle's dirent array with information about all entries.
	 */
	if (h->entry->size != 0) {

		h->desize = h->entry->size;

		h->de = (struct dirent *)
		    malloc(sizeof(struct dirent *) * h->desize);

		if (h->de != NULL) {
			bzero(h->de, sizeof(struct dirent *) * h->desize);

			de = h->de;
			PLIST_FOREACH(&h->entry->contents.childs, e, list) {
				plock_splock(&e->lock);
				strncpy(de->d_name, e->name, sizeof(de->d_name));
				plock_unlock(&e->lock);
				de++;
			}
	
			/*
			 * In the case of an opendir the handler's offset
			 * member is used as an index to traverse the dirent
			 * array (see plfs_readdir()).
			 */
			h->offset = 0;

		} else {
			st = PLSTATUS_FAILED;
		}
	}

	plock_unlock(&h->entry->lock);

	if (st == PLSTATUS_SUCCESS) {
		*handle = hnd;
	} else {
		plfs_close(hnd);
	}

	return st;
}


/**
 * Close a directory handle.
 *
 * @param handle (IN) Handle associated to the open entry.
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see plfs_close()
 */
static PLSTATUS
plfs_closedir(long handle)
{
	PLFS_HANDLE *h;

	h = (PLFS_HANDLE *)handle;

	plock_splock(&h->lock);

	if (h->de != NULL) {
		free(h->de);
		h->de = NULL; //XXX to avoid a race between locks
	}

	plock_unlock(&h->lock);

	return plfs_close(handle);
}


/**
 * Retrieve the next directory entry.
 *
 * @param handle (IN) Handle to the open directory
 * (returned by plfs_opendir());
 *
 * @param de (OUT) Updated with the next entry.
 *
 * @return A peludo status code.
 *
 * @see PLSTATUS
 * @see plfs_opendir()
 */
static PLSTATUS
plfs_readdir(long handle, struct dirent **de)
{
	PLFS_HANDLE *h;
	PLSTATUS st;

	h = (PLFS_HANDLE *)handle;

	plock_splock(&h->lock);

	/*
	 * Retrieve the next entry.
	 */
	if (h->offset < h->desize) {
		*de = h->de + h->offset;
		h->offset++;
		st = PLSTATUS_SUCCESS;
	} else {
		st = PLSTATUS_EOF;
	}

	plock_unlock(&h->lock);

	return st;
}


/**
 * Rewind the directory offset to the first element.
 *
 * @param handle (IN) Handle to the open directory
 * (returned by plfs_opendir()).
 *
 * @return A peludo status code.
 *
 * @see PLSTATUS
 * @see plfs_opendir()
 * @see plfs_readdir()
 */
static PLSTATUS
plfs_rewinddir(long handle)
{
	PLFS_HANDLE *h;

	h = (PLFS_HANDLE *)handle;

	plock_splock(&h->lock);
	h->offset = 0;
	plock_unlock(&h->lock);

	return PLSTATUS_SUCCESS;
}


/**
 * Create a new directory.
 *
 * @param path (IN) Fullpath of directory to create.
 *
 * @return A Peludo status code (notice that if the directory already
 * exists it will exit with an error condition).
 *
 * @see PLSTATUS
 * @see plfs_open()
 */
static PLSTATUS
plfs_mkdir(const char *path)
{
	PLSTATUS st;
	long handle;

	st = plfs_open(&handle, path, O_CREAT | O_EXCL | O_DIRECTORY, 0, 0);
	if (st == PLSTATUS_SUCCESS) {
		(void)plfs_close(handle);
	}

	return st;
}


/**
 * XXX
 * Release all PLFS resorces.
 * This function will fail if there is an open entry.
 */
static PLSTATUS
plfs_release(PLVFS *fs)
{
	return PLSTATUS_SUCCESS;
}


/**
 * Initialize the PLFS file system.
 * Initializes the plfs root directory and fill the
 * filesystem operations structure.
 *
 * @param fs (OUT) Updated with filesystem operation callbacks
 * (should not be <code>NULL</code>).
 *
 * @return A Peludo status code.
 *
 * @see PLSTATUS
 * @see PLVFS
 * @see plfs_root
 */
PLSTATUS PLEXPORT
plfs_init(PLVFS *fs)
{
	/*
	 * Initialize the root entry.
	 */
	PLIST_INIT(&plfs_root.list);
	PLIST_INIT(&plfs_root.contents.childs);
	plfs_root.isdir = 1;

	/*
	 * Set filesystem callbacks.
	 */
	fs->open = plfs_open;
	fs->close = plfs_close;
	fs->lseek = plfs_lseek;
	fs->fstat = plfs_fstat;
	fs->stat = plfs_stat;
	fs->ftruncate = plfs_ftruncate;
	fs->unlink = plfs_unlink;
	fs->rename = plfs_rename;
	fs->read = plfs_read;
	fs->write = plfs_write;

	fs->opendir = plfs_opendir;
	fs->closedir = plfs_closedir;
	fs->readdir = plfs_readdir;
	fs->rewinddir = plfs_rewinddir;
	fs->mkdir = plfs_mkdir;
	fs->rmdir = plfs_unlink;

	fs->serialize = NULL;
	fs->unserialize = NULL;

	fs->release = plfs_release;

	return PLSTATUS_SUCCESS;
}


/* XXX remove */
void PLEXPORT
plfs_getplfs(PLVFS *fs)
{
	/*
	 * Set filesystem callbacks.
	 */
	fs->open = plfs_open;
	fs->close = plfs_close;
	fs->lseek = plfs_lseek;
	fs->fstat = plfs_fstat;
	fs->stat = plfs_stat;
	fs->ftruncate = plfs_ftruncate;
	fs->unlink = plfs_unlink;
	fs->rename = plfs_rename;
	fs->read = plfs_read;
	fs->write = plfs_write;

	fs->opendir = plfs_opendir;
	fs->closedir = plfs_closedir;
	fs->readdir = plfs_readdir;
	fs->rewinddir = plfs_rewinddir;
	fs->mkdir = plfs_mkdir;
	fs->rmdir = plfs_unlink;

	fs->serialize = NULL;
	fs->unserialize = NULL;

	fs->release = plfs_release;
}


