/*
 * This test program is a command interpreter.
 * Commands are operations that you can execute inside a
 * simulated environment using a real PLFS filesystem.
 */
#include <plc.h>

#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <plfs.h>


PLVFS _plfs, *plfs = &_plfs;


static void cmd_ls(int argc, char **argv)
{
	struct dirent *de;
	long handle;
	PLSTATUS st;
	int i;

	for (i = 1; i < argc; i++) {
		printf("%s: <%s>\n", argv[0], argv[i]);
		st = plfs->opendir(&handle, argv[i]);
		if (st != PLSTATUS_SUCCESS) {
			printf("%s: Can't opendir <%s>: %d - %s\n",
			    argv[0],
			    argv[i],
			    st,
			    strerror(errno));
			return;
		}
		
		do {
			st = plfs->readdir(handle, &de);
			switch (st) {
			case PLSTATUS_SUCCESS:
				printf("%s/%s\n", argv[i], de->d_name);
			case PLSTATUS_EOF:
				break;
			default:
				printf("%s: Can't readdir: %d - %s\n",
				    argv[0],
				    st,
				    strerror(errno));
			}

		} while (st == PLSTATUS_SUCCESS);

		plfs->closedir(handle);
	}
}


static void cmd_rm(int argc, char **argv)
{
	while (--argc) {
		printf("%s: <%s>\n", argv[0], argv[argc]);
		if (plfs->unlink(argv[argc]) != PLSTATUS_SUCCESS) {
			printf("%s: Can't remove <%s>: %s\n\n",
			    argv[0], argv[argc], strerror(errno));
		}
	}
}


static void cmd_mkdir(int argc, char **argv)
{
	while (--argc) {
		printf("%s: <%s>\n", argv[0], argv[argc]);
		if (plfs->mkdir(argv[argc]) != PLSTATUS_SUCCESS) {
			printf("%s: Can't mkdir <%s>: %s\n",
			    argv[0], argv[argc], strerror(errno));
		}
	}
}


static void cmd_rmdir(int argc, char **argv)
{
	while (--argc) {
		printf("%s: <%s>\n", argv[0], argv[argc]);
		if (plfs->rmdir(argv[argc]) != PLSTATUS_SUCCESS) {
			printf("%s: Can't rmdir <%s>: %s\n",
			    argv[0], argv[argc], strerror(errno));
		}
	}
}


static void cmd_stat(int argc, char **argv)
{
	struct stat buf;
	PLSTATUS st;
	int i;

	for (i = 1; i < argc; i++) {
		st = plfs->stat(argv[i], &buf);
		if (st == PLSTATUS_SUCCESS) {
			printf("stat of %s: size=%d blocks=%d blocksize=%d\n",
			    argv[i],
			    buf.st_size,
			    buf.st_blocks,
			    buf.st_blksize);
		} else {
			printf("%s: Can't stat <%s>: %s\n",
			    argv[0], argv[i], strerror(errno));
		}
	}
}


static void cmd_cat(int argc, char **argv)
{
	char buf[10];
	size_t size;
	long handle;
	PLSTATUS st;
	int i, j;

	for (i = 1; i < argc; i++) {
		st = plfs->open(&handle, argv[i], 0, 0, 0);
		if (st != PLSTATUS_SUCCESS) {
			printf("%s: Can't open <%s>: %s\n",
			    argv[0], argv[i], strerror(errno));
			continue;
		} 

		do {
			st = plfs->read(handle, buf, sizeof(buf), &size);
			if (st != PLSTATUS_SUCCESS) {
				printf("%s: Can't read from <%s>: %s\n",
				    argv[0],
				    argv[i],
				    strerror(errno));
				break;
			}

			for (j = 0; j < size; j++) {
				printf("%c", buf[j]);
			}

		} while (size);

		plfs->close(handle);
	}

	printf("\n");
}


static void cmd_write(int argc, char **argv)
{
	long handle;
	int i;

	if (argc < 2) {
		return;
	}

	if (plfs->open(&handle, argv[1], O_CREAT | O_APPEND, 0, 0)
	    != PLSTATUS_SUCCESS) {
		printf("%s: Can't create <%s>: %s\n",
		    argv[0], argv[1], strerror(errno));
		return;
	}

	for (i = 2; i < argc; i++) {
		if (plfs->write(handle, argv[i], strlen(argv[i]), NULL)
		    != PLSTATUS_SUCCESS) {
			printf("%s: Can't write to <%s>: %s\n",
			    argv[0], argv[1], strerror(errno));
			break;
		}
		if (i + 1 < argc) {
			plfs->write(handle, " ", 1, NULL);
		}
	}

	plfs->close(handle);
}


static void cmd_writec(int argc, char **argv)
{
	long handle;
	int i;

	if (argc < 2) {
		return;
	}

	if (plfs->open(&handle, argv[1], O_CREAT | O_TRUNC, 0, 0)
	    != PLSTATUS_SUCCESS) {
		printf("%s: Can't create <%s>: %s\n",
		    argv[0], argv[1], strerror(errno));
		return;
	}

	for (i = 2; i < argc; i++) {
		if (plfs->write(handle, argv[i], strlen(argv[i]), NULL)
		    != PLSTATUS_SUCCESS) {
			printf("%s: Can't write to <%s>: %s\n",
			    argv[0], argv[1], strerror(errno));
			break;
		}
		if (i + 1 < argc) {
			plfs->write(handle, " ", 1, NULL);
		}
	}

	plfs->close(handle);
}


static void cmd_writeo(int argc, char **argv)
{
	long handle;
	off_t offset;
	int i;

	if (argc < 3) {
		return;
	}

	if (plfs->open(&handle, argv[1], O_CREAT | O_APPEND, 0, 0)
	    != PLSTATUS_SUCCESS) {
		printf("%s: Can't create <%s>: %s\n",
		    argv[0], argv[1], strerror(errno));
		return;
	}

	offset = (off_t)strtol(argv[2], NULL, 0);

	for (i = 3; i < argc; i++) {
		if (plfs->lseek(handle, offset, SEEK_SET, NULL)
		    != PLSTATUS_SUCCESS) {
			printf("%s: Can't lseek <%s>: %s\n",
			    argv[0], argv[1], strerror(errno));
			break;
		}
		if (plfs->write(handle, argv[i], strlen(argv[i]), NULL)
		    != PLSTATUS_SUCCESS) {
			printf("%s: Can't write to <%s>: %s\n",
			    argv[0], argv[1], strerror(errno));
			break;
		}
		if (i + 1 < argc) {
			plfs->write(handle, " ", 1, NULL);
		}
	}

	plfs->close(handle);
}


int finish = 0;

static void cmd_exit(int argc, char **argv)
{
	finish = 1;
}


static void cmd_help(int argc, char **argv);

struct {
	char *name;
	void (*callback)(int, char **);
	char *help;
} commands[] = {
	{ "ls",		cmd_ls,		"ls <path>" },
	{ "rm",		cmd_rm,		"rm <path>" },
	{ "mkdir",	cmd_mkdir,	"mkdir <path>" },
	{ "rmdir",	cmd_rmdir,	"rmdir <path>" },
	{ "stat",	cmd_stat,	"stat <path>" },
	{ "cat",	cmd_cat,	"cat <file>" },
	{ "write",	cmd_write,	"write <file> <msg> ;# append" },
	{ "writeo",	cmd_writeo,	"writeo <file> <offset> <msg> ;# from offset" },
	{ "writec",	cmd_writec,	"writec <file> <msg> ;# create/truncate" },
	{ "exit",	cmd_exit,	"exit" },
	{ "quit",	cmd_exit,	"quit" },
	{ "help",	cmd_help,	"help" },
	{ NULL, NULL, NULL }
};


static void cmd_help(int argc, char **argv)
{
	int i;

	for (i = 0; commands[i].name != NULL; i++) {
		printf("\t%s\n", commands[i].help);
	}
}


#if 0
static void pldir(const char *path)
{
	struct dirent *de;
	long handle;
	PLSTATUS st;

	st = plsys_opendir(&handle, path);
	if (st != PLSTATUS_SUCCESS) {
		plprint("Can't plsys_opendir: %s", plstrstatus(st));
		return;
	}

	while (1) {
		st = plsys_readdir(handle, &de);
		if (st == PLSTATUS_SUCCESS) {
			plprint("\t%s\n", de->d_name);

		} else if (st == PLSTATUS_EOF) {
			break;

		} else {
			plprint("Can't plsys_readdir: %s - errno = %d\n",
			    plstrstatus(st), errno);
			break;

		}
	}

	plsys_closedir(handle);
}


static void dir(const char *path)
{
	struct dirent *de;
	DIR *dir;

	dir = opendir(path);
	if (dir == NULL) {
		fprintf(stderr, "Can't opendir: %s", strerror(errno));
		return;
	}

	while ((de = readdir(dir)) != NULL) {
		fprintf(stdout, "\t%s\n", de->d_name);
	}

	closedir(dir);
}

#endif


void readline(char *buf, size_t bufsiz)
{
	char c;

	do {
		plsys_read(0, &c, 1, NULL);
		if (c == '\n') {
			c = 0;
		}
		*buf = c;
		buf++;
	} while (bufsiz-- && c);
}


void
plshell()
{
#define MAXARGV 100
	char buf[1024], *pbuf, *argv[MAXARGV], *p;
	int i, argc;

	while (!finish) {
		printf("plfsh> ");
		readline(buf, sizeof(buf));
		pbuf = buf;
		argc = 0;
		do {
			p = strsep(&pbuf, " ");
			argv[argc++] = p;
		} while (argc < MAXARGV && pbuf != NULL);
		
		for (i = 0; commands[i].name != NULL; i++) {
			if (!strcmp(commands[i].name, argv[0])) {
				commands[i].callback(argc, argv);
			}
		}
	}
}


int
main(int argc, char **argv)
{
	int i;

	printf("argc = %d\n", argc);
	for (i = 0; i < argc; i++) {
		printf("argv[%d] = \"%s\"\n", i, argv[i]);
	}

	if (plfs_init(&_plfs) == PLSTATUS_SUCCESS) {
		plshell();
	} else {
		printf("plfs_init failed\n");
	}

	return 0;
}


