
#include <plc.h>
#include <pld.h>
#include <pldhash.h>
#include <plspinlock.h>

#define LIBC_PLD	"libplc.pld"


/*
 * The profiler must import these functions manually when the libc is loaded.
 * see load().
 */
static void *libc = NULL;
static int dummy();
static int dummyl(long, ...);
static int dummyv(void *, ...);
static PLSTATUS	(*_plsys_gettimeofday)(struct timeval *, void *) =
		(void *)dummyv;
static void	(*_plsys_exit)(long) = (void *)dummyl; 
static PLSTATUS	(*_plsys_open)(long *, const char *, int, int) =
		(void *)dummyv; 
static void	(*_plsys_write)(long, void *, int, int *) = (void *)dummyl;
static long	(*_plsys_gettid)() = (void *)dummy;
static int	(*_plprint)(const char *, ...) = (void *)dummyv;
static int	(*_plsnprint)(char *, int, const char *, ...) = (void *)dummyv;

static struct { 
	char *name;
	void **val;
} nv[] = {
	{ "plsys_gettimeofday",	(void **)&_plsys_gettimeofday },
	{ "plsys_exit",		(void **)&_plsys_exit },
	{ "plsys_open",		(void **)&_plsys_open },
	{ "plsys_write",	(void **)&_plsys_write },
	{ "plsys_gettid",	(void **)&_plsys_gettid },
	{ "plprint",		(void **)&_plprint },
	{ "plsnprint",		(void **)&_plsnprint },
	{ NULL, NULL }
};


static long fd = -1;


/**
 * Dummy function.
 * Prevents unnecessary segmentation faults when something
 * goes wrong during the profiler auto-initialization.
 */
static int
dummy()
{
	asm("int3");
	return 0;
}


/**
 * Dummy function.
 * Prevents unnecessary segmentation faults when something
 * goes wrong during the profiler auto-initialization.
 */
static int
dummyv(void *v, ...)
{
	asm("int3");
	return 0;
}


/**
 * Dummy function.
 * Prevents unnecessary segmentation faults when something
 * goes wrong during the profiler auto-initialization.
 */
static int
dummyl(long l, ...)
{
	asm("int3");
	return 0;
}


/**
 * Get the length of a string.
 * @param str Asciiz string;
 * @return The length.
 */
static int
_strlen(const char *str)
{
	int i;

	for (i = 0; *str != '\0'; i++, str++);

	return i;
}


/**
 * Hash method that we hope it is used to generate PLD files.
 * @param buf Buffer;
 * @param len Length of the buffer.
 * @return A 32 bits hash.
 */
static uint32
_pldhash(const char *buf, unsigned int len)
{
	unsigned int hash = 5381, i = 0;

	for (i = 0; i < len; buf++, i++) {
		hash = ((hash << 5) + hash) + (*buf);
	}

	return ((uint32)hash);
}


/**
 * Resolve all profiler dependencies with the libc.
 * Since the profiler is loaded before the libc it cannot use any
 * libc symbols until it is loaded, for the same reason all used
 * methods can not be resolved by the peludo linker but manually.
 * @return A non-null value on success; otherwise <code>0</code>.
 */
static int
load()
{
	static PLOCK lock = PLOCK_INITIALIZED;
	int i;

	plock_splock(&lock);

	if (libc == NULL) {
		libc = pld_lookup(_pldhash(LIBC_PLD, sizeof(LIBC_PLD) - 1));
		if (libc == NULL) {
			plock_unlock(&lock);
			return 0;
		}

		/*
		 * Hope that all is resolved.
		 */
		for (i = 0; nv[i].name != NULL; i++) {
			pld_sym(nv[i].val, libc, _pldhash(nv[i].name,
			    _strlen(nv[i].name)));
		}

		/*
		 * Open the log file.
		 */
		if (_plsys_open(&fd, "profiler.txt",
		    O_RDWR | O_APPEND | O_CREAT, 0644) != PLSTATUS_SUCCESS) {
			_plprint("%s: Can't open: errno=%d\n",
			    __FUNCTION__, errno);
			plock_unlock(&lock);
			return 0;
		}
	}

	plock_unlock(&lock);
	return 1;
}


/**
 * Resolve a symbol.
 * @param pldname PLDNAME of the symbol to resolve.
 * @return The virtual address of the symbol or
 * <code>0</code> if it does not exist.
 * @see PLDNAME
 */
long __attribute__((stdcall))
prof_pldsym(const char *name, long pldname)
{
	PLSTATUS st;
	long sym;

	if (libc == NULL && !load()) {
		return 0;
	}

	if ((st = pld_sym((void **)&sym, libc, pldname)) != PLSTATUS_SUCCESS) {
		_plprint("%s: Can't pld_sym: status=%d, symbol=\"%s\", "
		    "pldname=0x%x\n",
		    __FUNCTION__,
		    st,
		    name,
		    pldname); 

		_plsys_exit(-1);
		return 0;
	}

	return sym;
}


/**
 * Return the number of current microseconds.
 */
uint64 __attribute__((stdcall))
prof_useconds()
{
	struct timeval tv;

	_plsys_gettimeofday(&tv, NULL);

	return (((uint64)tv.tv_sec << 32) | tv.tv_usec);
}


/**
 * Track a symbol.
 * An entry in the profiler file is added with the following data:
 * <pre><symbol> <thread id> <elapsed time in us></pre>
 * The <code>elapsed time</code> is the difference between the
 * <code>uend</code> and the <code>ustart</code> parameters.
 * @param symbol Symbol name;
 * @param ustart Starting microseconds;
 * @param uend Ending microseconds.
 */
void __attribute__((stdcall))
prof_tracker(const char *symbol, long starts, long startu, long ends,
    long endu)
{
	static PLOCK lock = PLOCK_INITIALIZED;
	uint64 start, end;
	char buf[1024];
	int len;

	start = starts * 1000000L + startu;
	end = ends * 1000000L + endu;

	len = _plsnprint(buf, sizeof(buf), "%s %d %d\n", symbol,
	    _plsys_gettid(), (uint32)(end - start));

	plock_splock(&lock);
	_plsys_write(fd, buf, len, NULL);
	plock_unlock(&lock);
}


