/**
 * @file src/lib/pthread/attr.c
 * Peludo Threads - Attributes.
 */
#include <plc.h>
#include <pllog.h>

#include <pthread.h>
#include <string.h>

#include "pthread_p.h"


/**
 * Allocate stack memory.
 *
 * @param attr (IN/OUT) Attributes where the desired stack size
 * is specified; it is updated with the new (page aligned) size
 * and the allocated stack address.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_attr_t
 * @see attr_stack_free()
 */
int 
attr_stack_alloc(pthread_attr_t *attr)
{
	PLSTATUS st;
	long err;

	if (attr->attr_stack_size == 0) {
		attr->attr_stack_addr = 0;
		return PLSTATUS_SUCCESS;
	}

	attr->attr_stack_size = PLALIGN(attr->attr_stack_size, PLPAGE_SIZE);

	st = _plsys_palloc(&err, (&attr->attr_stack_addr),
	    attr->attr_stack_size, PLMFLAGS_R | PLMFLAGS_W);

	if (st != PLSTATUS_SUCCESS) {
		pldebug("%s: Stack allocation failed: %s, errno=%d\n",
		    __FUNCTION__,
		    plstrstatus(st),
		    err);

		return ((int)err);

	} else {
		pldebug("%s: New stack=%p, size=%d (0x%x)\n",
		    __FUNCTION__,
		    attr->attr_stack_addr,
		    attr->attr_stack_size,
		    attr->attr_stack_size);
	}

	return 0;
}


/**
 * Release all resources used by an attribute object.
 *
 * @param attr (IN) Attributes object to release.
 *
 * @return <code>0</code> is returned.
 *
 * @see pthread_attr_t
 * @see pthread_attr_init()
 */
int PLEXPORT
pthread_attr_destroy(pthread_attr_t *attr)
{
	if (attr != NULL &&
	    attr->attr_stack_size != 0 &&
	    attr->attr_stack_addr != 0) {

		pldebug("%s: Releasing stack=%p, size=%d (0x%x)\n",
		    __FUNCTION__,
		    attr->attr_stack_addr,
		    attr->attr_stack_size,
		    attr->attr_stack_size);

		plsys_pfree(attr->attr_stack_addr, attr->attr_stack_size);

		attr->attr_stack_addr = 0;
	}

	return 0;
}


/**
 * Initialize a thread attribute object with default values.
 * Default values are:
 *   - Detached thread state;
 *   - <code>_PTHREAD_DEFAULT_STACK_SIZE_NP</code> stack size.
 *
 * @param attr (IN) Attributes to initialize.
 *
 * @return <code>0</code> is returned.
 *
 * @see _PTHREAD_DEFAULT_STACK_SIZE
 * @see pthread_attr_t
 * @see pthread_attr_destroy()
 */             
int PLEXPORT
pthread_attr_init(pthread_attr_t *attr)
{
	if (attr != NULL) {
		attr->attr_state = PTHREAD_CREATE_DETACHED;
		attr->attr_stack_size = PTHREAD_DEFAULT_STACK_SIZE_NP;
		attr->attr_stack_addr = 0;
	}
	return 0;
}


/**
 * Get the stack address attribute.
 *
 * @param attr (IN) Attributes containing the stack parameters;
 *
 * @param addr (OUT) If not <code>NULL</code> updated with the stack address.
 *
 * @return <code>0</code> is returned.
 *
 * @see pthread_attr_t
 * @see pthread_attr_getstacksize()
 * @see pthread_attr_init()
 */              
int PLEXPORT
pthread_attr_getstackaddr(const pthread_attr_t *attr, void **addr)
{
	if (attr != NULL && addr != NULL) {
		*addr = (void *)attr->attr_stack_addr;
	}
	return 0;
}


/**
 * Get the stack size attribute.
 *
 * @param attr (IN) Attributes containing the stack parameters;
 *
 * @param size (OUT) If not <code>NULL</code> updated with the stack size.
 *
 * @return <code>0</code> is returned.
 *
 * @see pthread_attr_t
 * @see pthread_attr_init()
 * @see pthread_attr_getstackaddr()
 * @see pthread_attr_setstacksize()
 */              
int PLEXPORT
pthread_attr_getstacksize(const pthread_attr_t *attr, size_t *size)
{
	if (attr != NULL && size != NULL) {
		*size = attr->attr_stack_size;
	}
	return 0;
}


/**
 * Set the stack size attribute.
 *
 * @param attr (IN/OUT) If not <code>NULL</code>
 * updated with the desired stack size;
 *
 * @param size (IN) Stack size.
 *
 * @return <code>0</code> is returned.
 *
 * @see pthread_attr_t
 * @see pthread_attr_init()
 * @see pthread_attr_getstacksize()
 */            
int PLEXPORT
pthread_attr_setstacksize(pthread_attr_t *attr, size_t size)
{
	if (attr != NULL) {
		attr->attr_stack_size = size;
	}
	return 0;
}


/**
 * Get the detach state attribute.
 *
 * @param attr (IN/OUT) If not <code>NULL</code>
 * updated with the desired stack size;
 *
 * @param state (OUT) Updated with the detach state.
 *
 * @return <code>0</code> is returned.
 *
 * @see PTHREAD_CREATE_DETACHED
 * @see PTHREAD_CREATE_JOINABLE
 * @see pthread_attr_t
 * @see pthread_attr_init()
 * @see pthread_attr_setdetachstate()
 */            
int PLEXPORT
pthread_attr_getdetachstate(pthread_attr_t *attr, int *state)
{
	if (attr != NULL && state != NULL) {
		*state = attr->attr_state;
	}
	return 0;
}


/**
 * Set the detach state attribute.
 *
 * @param attr (IN/OUT) If not <code>NULL</code>
 * updated with the desired stack size;
 *
 * @param state (IN) Detach state to set.
 *
 * @return <code>0</code> is returned.
 *
 * @see PTHREAD_CREATE_DETACHED
 * @see PTHREAD_CREATE_JOINABLE
 * @see pthread_attr_t
 * @see pthread_attr_init()
 * @see pthread_attr_getdetachstate()
 */            
int PLEXPORT
pthread_attr_setdetachstate(pthread_attr_t *attr, int state)
{
	if (attr != NULL) {
		attr->attr_state = state;
	}
	return 0;
}


/**
 * Get a copy of the attributes of a specified thread.
 *
 * @param th (IN) Thread whose attributes are wanted;
 *
 * @param attr (OUT) updated with the thread's attributes.
 *
 * @return <code>0</code>.
 */
int PLEXPORT
pthread_attr_get_np(pthread_t th, pthread_attr_t *attr)
{
	if (attr != NULL) {
		memcpy(attr, &th->thread_attr, sizeof(pthread_attr_t));
	}

	return 0;
}


int PLEXPORT
pthread_attr_setscope(pthread_attr_t *attr, int contentionscope)
{
	return ENOSYS;
}


