/**
 * @file src/lib/plt/mutex.c
 * Peludo Threads - Mutex.
 */
#include <plc.h>
#include <pllog.h>

#include <pthread.h>

#include "pthread_p.h"


/**
 * Initialize a Mutex.
 * The mutex is initialized unlocked.
 *
 * @param mutex Mutex to initialize.
 *
 * @return <code>0</code>.
 *
 * @see pthread_mutex_lock()
 * @see pthread_mutex_unlock()
 *
 * TODO Mutex attributes.
 */
int PLEXPORT
pthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutex_attr_t *mattr)
{
	if (mattr != NULL) {
		pldebug("%s: Mutex attributes not supported. Aborting.\n",
		    __FUNCTION__);
		plsys_exit(-1);
		/* NOTREACHED */
	}

	if (mutex != NULL) {
		mutex->mtx_lock = 0;
		mutex->mtx_count = 0;
	}

	return 0;
}


/**
 * Destroy an initialized and unlocked mutex.
 *
 * @param mutex Mutex to destroy.
 *
 * @return <code>0</code>.
 *
 * @see pthread_mutex_init()
 * @see pthread_mutex_unlock()
 */
int PLEXPORT
pthread_mutex_destroy(pthread_mutex_t *mutex)
{
	return 0;
}


/**
 * Lock a Mutex.
 * <li>Peludo Mutexes are recursive, so when a thread tries to lock an
 * already owned mutex an internal counter is incremented;
 * to completely unlock that mutex an equal number of unlock
 * operations must be done.</li>
 * <li>Peludo Mutexes are thread and signal safe: You can lock and unlock
 * mutexes from a signal handler.</li>
 *
 * @param mutex Mutex to lock;
 *
 * @param block A non-zero value means that the current thread will block
 * (put to sleep) until the mutex is owned; otherwise it will just try to own
 * the mutex and return.
 *
 * @return <code>0</code> if the mutex is locked;
 * if the mutex is already owned and <code>block</code> is zero
 * <code>EBUSY</code> is returned; on error another code is returned.
 *
 * @see phtread_mutex_unlock()
 */
int PLEXPORT
pthread_mutex_lock_np(pthread_mutex_t *mutex, int block)
{
	long tid, cur;
//XXX	sigset_t set;
	PLSTATUS st;

	if (mutex == NULL) {
		return EINVAL;
	}

	tid = plsys_gettid();

	while (1) {

//XXX		plsigblock(&set);

		/*
		 * Try to own the mutex.
		 */
		cur = platomic_iex((long *)&mutex->mtx_lock, 0, tid);
		if (cur == 0 || cur == tid) {
			/*
			 * Owned.
			 */
			mutex->mtx_count++;

//XXX			plsigunblock(&set);

			return 0;
		}

//XXX		plsigunblock(&set);

		if (!block) {
			/*
			 * Try again later.
			 */
			break;
		}

		/*
		 * Sleep until the owner unlocks the mutex.
		 */
		st = plsys_twait((long *)&mutex->mtx_lock, cur, -1);
		switch (st) {
		case PLSTATUS_SUCCESS:
		case PLSTATUS_TRY_AGAIN:
			break;

		default:
			/*
			 * Error.
			 */
			return errno;
		}
	}

	/*
	 * Tried but locked by another thread.
	 */
	return EBUSY;
}

/**
 * Lock a Mutex.
 *
 * @param mutex Mutex to lock.
 *
 * @return <code>0</code> on success <code>0</code>;
 * otherwise an error code.
 *
 * @see plt_mutex_lock_np()
 */
int PLEXPORT
pthread_mutex_lock(pthread_mutex_t *mutex)
{
	return pthread_mutex_lock_np(mutex, 1);
}


/**
 * Try to lock a Mutex.
 *
 * @param mutex Mutex to try to lock;
 *
 * @return If the mutex is locked <code>0</code> is returned;
 * if the mutex is already owned by another thread <code>EBUSY</code>
 * is returned; on error another code is returned.
 *
 * @see plt_mutex_lock_np()
 */
int PLEXPORT
pthread_mutex_trylock(pthread_mutex_t *mutex)
{
	return pthread_mutex_lock_np(mutex, 0);
}


/**
 * Unlock a Mutex.
 * If the calling thread has locked the mutex multiple times an internal
 * counter is decremented; the mutex will be effectively unlocked when
 * this counter reaches 0.
 *
 * @param mutex (IN) Mutex to unlock.
 *
 * @return On success <code>0</code> is returned;
 * otherwise the error code is returned.
 *
 * @see pthread_mutex_lock_np()
 */
int PLEXPORT
pthread_mutex_unlock(pthread_mutex_t *mutex)
{
	long tid, cur;
//XXX	sigset_t set;
	int err;

	if (mutex == NULL) {
		return EINVAL;
	}

	tid = plsys_gettid();

//XXX	plsigblock(&set);

	/*
	 * Unlock only when mutex is owned by the calling thread.
	 */
	cur = platomic_iex((long *)&mutex->mtx_lock, tid, tid);
	if (cur == tid) {
		/*
 		 * The calling thread owns the mutex.
 		 */
		mutex->mtx_count--;

		if (mutex->mtx_count == 0) {
			/*
			 * Unlock and wakeup some waiting thread.
			 */
			platomic_iex((long *)&mutex->mtx_lock, tid, 0);
			err = (plsys_tsignal((long *)&mutex->mtx_lock, 1) >= 0);
		}
	} else {
		/*
 		 * Owned by another thread.
 		 */
		err = EPERM;
	}

//XXX	plsigunblock(&set);

	return err;
}


/**
 * Get the thread id of the thread that owns a specified mutex.
 *
 * @param mutex (IN) Mutex;
 *
 * @return If the mutex has an owner its thread id is returned;
 * otherwise <code>0</code> is returned.
 *
 * @see pthread_mutex_lock_np()
 * @see pthread_mutex_unlock()
 */
long PLEXPORT
pthread_mutex_ownertid_np(pthread_mutex_t *mutex)
{
	if (mutex == NULL) {
		return 0;
	}

	return platomic_iex((long *)&mutex->mtx_lock, 0, 0);
}


