/*
 * @file src/lib/pthread/pthread.c
 * Peludo Threads.
 */
#include <plc.h>
#include <pllog.h>

#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "pthread_p.h"


extern void (*plprepare_fork)();
extern void (*plparent_fork)();
extern void (*plchild_fork)();


/**
 * Force enable and asynchronous cancellation of threads.
 * This variable is set to 1 when the pthread library is
 * unloading and all threads must be terminated.
 *
 * @see _pthread_sighandler()
 * @see pthread_release_np()
 */
static int _pthread_cancel_forced = 0;


/**
 * Main thread's context.
 */
pthread_t _pthread_main = NULL;


/**
 * List of threads.
 */
PLIST _pthread_list = PLIST_INITIALIZED;
PLOCK _pthread_list_lock = PLOCK_INITIALIZED;


/**
 * Number of threads.
 */
static long _pthread_threads = 0;


/**
 * Previous (saved) errno locator.
 */
static int *(*_saved_geterrno)() = NULL;


/**
 * Previous (saved) termination callback.
 */
static void (*_previous_plexit)(long) = NULL;


/**
 * Previous (saved) cancellation enter point.
 */
static void (*_previous_entercancel)() = NULL;


/**
 * Previous (saved) cancellation leave point.
 */
static void (*_previous_leavecancel)() = NULL;


/**
 * Child fork callback.
 * This callback is called by fork(2) in the context of the new child,
 * it re-initializes the libc and pthreads internal parameters for the
 * new process.
 *
 * @see fork()
 */
static void
_child_fork()
{
	plinit();
	pthread_init_np();
}


/**
 * Termination callback.
 * This function is called by <code>plexit()</code> when a thread exits.
 * If the thread who called <code>plexit()</code> is a child it is
 * terminated through <code>pthread_exit()</code>; if it is the main thread
 * the previous saved termination callback is executed instead.
 *
 * @param retval (IN) Return value.
 *
 * @see _previous_plexit()
 * @see plexit()
 */
static void
_pthread_exit(long retval)
{
	if (plsys_gettid() != _pthread_main->thread_tid) {
		/*
		 * A child thread is terminated.
		 */
		pthread_exit((void *)retval);
		/* NOTREACHED */
	}

	/*
	 * The main thread is terminated.
	 */
	if (_previous_plexit != NULL) {
		_previous_plexit(retval);
	}

	/* If we are here we are returning to the caller: plexit(). */
}


/**
 * Add a thread inside the thread list.
 *
 * @param th (IN) Thread to add;
 *
 * @param tid (IN) Thread Id of the caller.
 *
 * @see pthread_t
 * @see _pthread_list
 * @see _pthread_list_lock
 */
static void
_pthread_add(const pthread_t th, long tid)
{
	sigset_t cset;

	plsigblock(&cset);
	plock_spvalock(&_pthread_list_lock, tid);

	plist_add_tail(&_pthread_list, (PLIST *)th);
	platomic_inc(&_pthread_threads);

	plock_unvalock(&_pthread_list_lock, tid);
	plsigunblock(&cset);
}


/**
 * Threaded errno locator.
 *
 * @return Return A pointer to the current thread's errno variable.
 */
static int *
_pthread_errno()
{
	pthread_t self;

	self = pthread_self();
	if (self == NULL) {
		/* Not a peludo thread */
		plsys_exit(-1);
		/* NOTREACHED */
	}

	/* No need to lock the _pthread structure */
	return (&self->thread_errno);
}


/**
 * Release resources of defunct threads.
 */
static void
_pthread_dead_collector()
{
	pthread_t th, main;
	long defuncts, tid;
	sigset_t set;

	/*
	 * Give some time to the dying threads to effectively die.
	 */
	plsys_schedyield();

	defuncts = 0;
	tid = plsys_gettid();

	/* XXX already blocked on the caller: signal handler */
	plsigblock(&set);
	plock_spvalock(&_pthread_list_lock, tid);

	main = PLIST_FIRST(&_pthread_list, pthread_t);

	/* XXX Nasty loops */
	do {
		PLIST_FOREACH_REV(&_pthread_list, th, thread_list) {
			if (th == main) {
				/*
				 * There are no more child threads.
				 */
				goto bye;
			}

			plock_spvalock(&th->thread_lock, tid);

			/*
			 * Release resources of defunct threads.
			 */
			if (platomic_iex(&th->thread_state, PTHREAD_STATE_DEAD,
			    PTHREAD_STATE_DEAD) == PTHREAD_STATE_DEAD) {
				/*
				 * Release stack and TLS.
				 */
				pthread_attr_destroy(&th->thread_attr);
				tls_destructor(th);
	
				/*
				 * Remove from the thread list.
				 */
				plist_del((PLIST *)th);
	
				/*
				 * Release context.
				 */
				free(th);
	
				defuncts++;

				platomic_dec(&_pthread_threads);

				/*
				 * Because of the way lists works
				 * we are forced to break and start again.
				 */
				break;
			}

			plock_unvalock(&th->thread_lock, tid);
		}

	} while (defuncts);

bye:	pldebug("%s: Released %d defunct threads.\n",
	    __FUNCTION__,
	    defuncts);

	plock_unvalock(&_pthread_list_lock, tid);
	plsigunblock(&set);
}


/**
 * Thread's signal handler.
 * This function handles two (real time) signals:
 *
 * - <code>SIGSETXID</code>: This signal is intended to be received by
 *   the main thread when a child is exiting. When it is received
 *   <code>_pthread_dead_collector()</code> is called in order to release
 *   resources of daed threads.
 *
 * - <code>SIGCANCEL</code>: This signal is intended to be received by
 *   any thread that should be terminated (it is like the
 *   <code>SIGTERM</code> signal but for threads). When it is received and
 *   cancellation is enabled and asynchornous the <code>plexit()</code> is
 *   called in order to terminate the thread; otherwise the
 *   <code>PTHREAD_STATE_CANCELED</code> state is set.
 *
 * - Other signals are ignored.
 *
 * @param signo (IN) Signal number.
 *
 * @see _pthread_dead_collector()
 * @see plexit()
 */
static void
_pthread_sighandler(int signo)
{
	int saved_errno;
	pthread_t self;
	sigset_t set;

	plsigblock(&set);

	saved_errno = errno;

	switch (signo) {
	case SIGSETXID:
		/*
		 * Release resources of defunct threads.
		 */
		_pthread_dead_collector();
		break;

	case SIGCANCEL:
		if (_pthread_cancel_forced) {
			/*
			 * Cancel is forced:
			 * The thread must be terminated now!
			 */
bye:			plexit(-1);
			/* NOTREACHED */
		}

		self = pthread_self();
		if (self->thread_cancelstate == PTHREAD_CANCEL_ENABLE &&
		    self->thread_canceltype == PTHREAD_CANCEL_ASYNCHRONOUS) {
			/*
			 * Asynchronous cancellation.
			 */
			goto bye;
		}

		/*
 		 * Cancel is disabled and/or deferred:
 		 * It will be killed at any cancellation point
 		 * when it becomes enabled.
 		 */
		self->thread_state = PTHREAD_STATE_CANCELED;
		break;

	default:
		pldebug("%s: Received signal=%d. Ignored.\n",
		    __FUNCTION__,
		    signo);
	}

	errno = saved_errno;

	plsigunblock(&set);
}


/**
 * Thread Starter Function.
 * This function is called when a new thread is created (using
 * <code<>pthread_create()</code>) and blocks until its context is unlocked;
 * it then calls its start routine; if that routine returns the thread
 * terminates itself by calling <code>pthread_exit()</code>.
 *
 * @param arg (IN) Thread's context.
 *
 * @see pthread_create()
 * @see pthread_exit()
 */
static void
_pthread_start(long arg)
{
	pthread_t th;
	void *retval;

	th = (pthread_t)arg;

	pldebug("%s: New thread: fn=%x, arg=%x. Synchronizing with parent...\n",
	    __FUNCTION__,
	    th->thread_start_fn,
	    th->thread_start_arg);

	/*
	 * Block until all context members are correctly
	 * initialized by the parent (see pthread_create()).
	 */
	plock_splock(&th->thread_lock);
	plock_unlock(&th->thread_lock);

	/* Let the kernel put a 0 into thread_state when the thread is dead */
	/* XXX move this inside plsys_tfork() */
	plsyscall(SYS_set_tid_address, &th->thread_state, 0, 0, 0, 0, 0);

	pldebug("%s: Starting thread...\n", __FUNCTION__);

	retval = th->thread_start_fn(th->thread_start_arg);

	pldebug("%s: Thread returned, exit code=%d. Terminating.\n",
	    __FUNCTION__,
	    retval);

	pthread_exit(retval);
	/* NOTREACHED */
}


/**
 * Initialize the pthread library.
 * Initialize internal components and create the main thread context.
 * <br>
 * This function should ONLY be called at initialization time when the
 * pthread library is statically linked to the application.
 * <br>
 * If the pthread library is dyanmically linked this function should not
 * be called (see <code>src/lib/pthread/init.c</code>).
 *
 * @return A Peludo Status code.
 *
 * @see PLSTATUS
 * @see pthread_release_np()
 * @see tls_init()
 */
PLSTATUS PLEXPORT
pthread_init_np()
{
	struct sigaction act;
	pthread_t th;
	PLSTATUS st;

	/*
	 * Initialize the TLS keys.
	 */
	if ((st = tls_init()) != PLSTATUS_SUCCESS) {
		return st;
	}

	/*
	 * Create the context of the main thread.
	 */
	th = (pthread_t)malloc(sizeof(struct _pthread));
	if (th == NULL) {
		tls_release();
		return PLSTATUS_OUT_OF_MEMORY;
	}

	/* 
	 * Initialize the list of threads.
	 */
	PLIST_INIT(&_pthread_list);

	/*
	 * Set the threaded errno locator.
	 */
	_saved_geterrno = plgeterrno;
	plseterrno(_pthread_errno);

	/*
	 * Set the cancellation point callbacks.
	 */
	plsetcancel(pthread_testcancel, pthread_testcancel,
	    &_previous_entercancel, &_previous_leavecancel);

	/*
	 * Set the callback to be called by the child after a fork().
	 */
	plchild_fork = _child_fork;

	/*
	 * Initialize the main thread's context.
	 */
	bzero((void *)th, sizeof(struct _pthread));

	pthread_attr_init(&th->thread_attr);
	pthread_attr_setstacksize(&th->thread_attr, 0);

	th->thread_pid = plsys_getpid();
	th->thread_tid = plsys_gettid();
	th->thread_state = PTHREAD_STATE_RUNNING;
	th->thread_cancelstate = PTHREAD_CANCEL_ENABLE;
	th->thread_canceltype = PTHREAD_CANCEL_DEFERRED;

	PLOCK_INIT_UNLOCKED(&th->thread_lock);
	PLIST_INIT(&th->thread_tlslist);
	PLIST_INIT(&th->thread_cleanuplist);

	/*
	 * Add the main thread's context into the thread list.
	 */
	_pthread_main = th;
	_pthread_add(th, th->thread_tid);

	/*
	 * Set the pthread termination callback (to be called by plexit()).
	 */
	plsetexit(_pthread_exit, &_previous_plexit);

	/*
	 * Set the pthread internal signal handler.
	 */
	bzero(&act, sizeof(act));
	act.sa_handler = _pthread_sighandler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGCANCEL, &act, NULL);
	sigaction(SIGSETXID, &act, NULL);

	return PLSTATUS_SUCCESS;
}


/**
 * Release the pthread library.
 * All threads are destroyed and resources released.
 * <br>
 * This function should ONLY be called at termination time when
 * the pthread library is statically linked to the application.
 * <br>
 * If the pthread library is dyanmically linked this function should
 * not be called (see <code>src/lib/pthread/init.c</code>).
 * <br>
 * <b>The caller of this functions should be the main thread</b>.
 *
 * @see _pthread_dead_collector()
 * @see pthread_init_np()
 * @see tls_release()
 */
void PLEXPORT
pthread_release_np()
{
	/*
	 * Be nice and send the termination signal to all threads.
	 */
	pldebug("%s: Sending termination signal to all threads...\n",
	    __FUNCTION__);

	/*
	 * Force termination of all threads.
	 */
	_pthread_cancel_forced = 1;
	pthread_killall_np(SIGCANCEL);

	/*
	 * Wait for all threads to exit nicely.
	 */
	do {
		plsys_msleep(50);
	} while (_pthread_threads > 1);

	/*
	 * Release the main thread.
	 */
	pthread_attr_destroy(&_pthread_main->thread_attr);
	tls_destructor(_pthread_main);
	free(_pthread_main);

	/*
	 * Release the TLS key array.
	 */
	tls_release();

	/*
	 * Restore previous cancellation callbacks.
	 */
	plsetcancel(_previous_entercancel, _previous_leavecancel,
	    NULL, NULL);

	/*
	 * Restore the previous errno locator.
	 */
	plseterrno(_saved_geterrno);
}


/**
 * Create a new thread.
 *
 * @param th (OUT) If not NULL it is updated with the context of
 * the new thread;
 *
 * @param attr (IN) Thread attributes (can be reused by the caller);
 * if this parameter is <code>NULL</code> the thread will be created
 * using the default attribute values (see pthread_attr_init());
 *
 * @param start (IN) Start routine;
 *
 * @param arg (IN) Argument to pass to the start routine.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_attr_t
 * @see pthread_t
 * @see _pthread_start()
 * @see pthread_attr_init()
 */
int PLEXPORT
pthread_create(pthread_t *th, pthread_attr_t *attr, void * (*start)(void *),
    void *arg)
{
	long stack, err;
	pthread_t newth;
	PLSTATUS st;

	/*
	 * Create the new thread's context.
	 */
	if ((newth = (pthread_t)malloc(sizeof(struct _pthread))) == NULL) {
		return errno;
	}

	bzero(newth, sizeof(struct _pthread));

	if (attr == NULL) {
		/*
		 * Default attributes.
		 */
		pthread_attr_init(&newth->thread_attr);
	} else {
		/*
		 * User defined attributes.
		 */
		newth->thread_attr = *attr;
	}

	err = attr_stack_alloc(&newth->thread_attr);
	if (err != 0) {
		goto fail2;
	}

#ifdef PLSTACK_DIRECTION_DOWN
	stack = newth->thread_attr.attr_stack_addr +
	    newth->thread_attr.attr_stack_size - sizeof(long); /* XXX */
#else
	stack = newth->thread_attr.attr_stack_addr;
#endif

	newth->thread_start_fn = start;
	newth->thread_start_arg = arg;
	newth->thread_pid = plsys_getpid();
	newth->thread_state = PTHREAD_STATE_RUNNING;
	newth->thread_cancelstate = PTHREAD_CANCEL_ENABLE;
	newth->thread_canceltype = PTHREAD_CANCEL_DEFERRED;
	PLOCK_INIT_UNLOCKED(&newth->thread_lock);
	PLIST_INIT(&newth->thread_tlslist);
	PLIST_INIT(&newth->thread_cleanuplist);

	/*
	 * Lock the context until all members are initialized
	 * (see _pthread_start()).
	 */
	plock_splock(&newth->thread_lock);

	/*
	 * Create the new thread (that will block until its context
	 * is unlocked, see _pthread_start()).
	 */
	st = plsys_tfork(&newth->thread_tid, stack, _pthread_start,
	    (long)newth);

	if (st != PLSTATUS_SUCCESS) {
		err = errno;
		pldebug("%s: Thread fork failed: peludo status=<%s>, "
		    "errno=%d\n",
		    __FUNCTION__,
		    plstrstatus(st),
		    err);
		goto fail;
	}

	pldebug("%s: New thread created: context=%p, tid=%d, stack=%p\n",
	    __FUNCTION__,
	    newth,
	    newth->thread_tid,
	    stack);

	/*
	 * Add the new thread into the thread list.
	 */
	_pthread_add(newth, plsys_gettid());

	/*
 	 * Start the new thread by unlocking its context (see _pthread_start()).
 	 */
	plock_unlock(&newth->thread_lock);

	if (th != NULL) {
		*th = newth;
	}

	return 0;

fail:	pthread_attr_destroy(&newth->thread_attr);

fail2:	free(newth);
	return err;
}


/**
 * Terminate the current thread.
 * If this function is called by the main thread the <code>plexit()</code>
 * function is executed forcing the termination of the Peludo System
 * (see <code>_pthread_exit()</code>).
 *
 * @param retval (IN) Return value.
 *
 * @see _pthread_exit()
 * @see plexit()
 * @see pthread_create()
 */
void PLEXPORT
pthread_exit(void *retval)
{
	pthread_t th;
	sigset_t set;

	plsigblock(&set);
	th = pthread_self();
	if (th == NULL) {
		/*
		 * Not a Peludo thread. Terminate it anyways.
		 */
		pldebug("%s: Terminating alien thread %d.\n",
		    __FUNCTION__,
		    plsys_gettid());

		plsys_texit((long)retval);
		/* NOTREACHED */
	}

	/*
	 * Execute cleanup callbacks.
	 */
	while (pthread_cleanup_pop_np(1));

	/*
	 * The main thread should call plexit().
	 */
	if (th == _pthread_main) {
		pldebug("%s: Terminating the Main thread, retval=%d...\n",
		    __FUNCTION__,
		    (long)retval);

		plsigunblock(&set);
		plexit((long)retval);
		/* NOTREACHED */
	}

	pldebug("%s: Terminating peludo thread %p, tid=%d, retval=%d\n",
	    __FUNCTION__,
	    th,
	    th->thread_tid,
	    retval);

	th->thread_retval = retval;

	/* Done by the kernel */
	//th->thread_state = PTHREAD_STATE_DEAD;

	/*
	 * Wake up joined threads.
	 * XXX Not implemented
	 * (joined threads must be waiting on the thread_alive futex, btw)
	 */

	/*
	 * Tell to the main thread that a child is terminating.
	 */
	pthread_kill(_pthread_main, SIGSETXID);

	/*
	 * Terminate.
	 */
	plsys_texit((long)retval);
	/* NOTREACHED */
}


/**
 * Get the context of a thread.
 *
 * @param tid (IN) Thread Id.
 *
 * @return If the given thread id corresponds to a peludo thread its
 * context is returned; otherwise <code>NULL</code> is returned.
 *
 * @see pthread_t
 */
pthread_t PLEXPORT
pthread_lookup_np(long tid)
{
	sigset_t cset;
	pthread_t th;
	long me;

	me = plsys_gettid();

	plsigblock(&cset);
	plock_spvalock(&_pthread_list_lock, me);

	PLIST_FOREACH(&_pthread_list, th, thread_list) {
		if (th->thread_tid == tid) {
			plock_unvalock(&_pthread_list_lock, me);
			plsigunblock(&cset);
			return th;
		}
	}

	plock_unvalock(&_pthread_list_lock, me);
	plsigunblock(&cset);

	return NULL;
}


/**
 * Get the context of the calling thread.
 *
 * @return If the calling thread is a peludo thread a pointer to
 * its context is returned; otherwise <code>NULL</code> is returned.
 *
 * @see pthread_t
 * @see pthread_lookup_np()
 */
pthread_t PLEXPORT
pthread_self()
{
#ifdef PLDEBUG
	pthread_t th;

	th = pthread_lookup_np(plsys_gettid());
	if (th == NULL) {
		pldebug("%s: Not a Peludo Thread\n", __FUNCTION__);
	}

	return th;

#else
	return pthread_lookup_np(plsys_gettid());

#endif
}


/* XXX */
int PLEXPORT
pthread_setschedparam(pthread_t th, int policy, const struct sched_param *sc)
{
	return ENOSYS;
}


