/**
 * @file src/lib/pthread/semaphore.c
 * Peludo Threads - Semaphores.
 */
#include <plc.h>
#include <pllog.h>

#include <stdlib.h>
#include <semaphore.h>


sem_t PLEXPORT *
sem_open(const char *name, int oflag, ...)
{
	errno = ENOSYS;
	return SEM_FAILED;
}


int PLEXPORT
sem_close(sem_t *sem)
{
	errno = ENOSYS;
	return -1;
}


int PLEXPORT
sem_unlink(const char *name)
{
	errno = ENOSYS;
	return -1;
}


/**
 * Initialize an unnamed semaphore object.
 *
 * @param sem (IN) Semaphore to initialize;
 *
 * @param pshare (IN) Must be <code>0</code> always;
 *
 * @param value (IN) Value used to initialize the semaphore.
 *
 * @return <code>0</code> on success; <code>-1</code> on error
 * and <code>errno</code> is set properly.
 *
 * @see sem_destroy()
 */
int PLEXPORT
sem_init(sem_t *sem, int pshared, unsigned int value)
{
	sem_t newsem;

	if (sem == NULL || ((long)value) >= SEM_VALUE_MAX) {
		errno = EINVAL;
		return -1;
	}

	if (pshared) {
		pldebug("%s: Shared semaphores are not supported\n");
		errno = EPERM;
		return -1;
	}

	newsem = (sem_t)malloc(sizeof(struct _sem));
	if (newsem == NULL) {
		errno = ENOSPC;
		return -1;
	}

	PLOCK_INIT_UNLOCKED(&newsem->sem_lock);
	newsem->sem_wcount = 0;
	newsem->sem_value = value;

	*sem = newsem;

	return 0;
}


/**
 * Destroy a semaphore.
 *
 * @param sem (IN) Semaphore to destroy.
 *
 * @return <code>0</code> on success; on error <code>errno</code>
 * is set and <code>-1</code> is returned.
 *
 * @see sem_init()
 */
int PLEXPORT
sem_destroy(sem_t *sem)
{
	if (sem == NULL) {
		errno = EINVAL;
		return -1;
	}

	if ((*sem)->sem_wcount) {
		pldebug("%s: Semaphore has waiters\n", __FUNCTION__);
		errno = EBUSY;
		return -1;
	}

	free(*sem);
	*sem = NULL;

	return 0;
}


/**
 * Get the current value of a semaphore.
 *
 * @param sem (IN) Semaphore;
 *
 * @param value (OUT) Updated with the current value of the semaphore.
 *
 * @return <code>0</code> on success; <code>-1</code> on error
 * (<code>errno</code> is set properly).
 */
int PLEXPORT
sem_getvalue(sem_t *sem, int *value)
{
	if (sem == NULL) {
		errno = EINVAL;
		return -1;
	}

	if (value != NULL) {
		*value = (*sem)->sem_value;
	}

	return 0;
}


/**
 * Increment (unlock) a semaphore.
 *
 * @param sem (IN) Semapthore to unlock.
 *
 * @return <code>0</code> on success; <code>-1</code> on error
 * (<code>errno</code> is set properly).
 *
 * XXX this function should be signal-safe (reentrant). Is it?
 *
 * @see sem_trywait()
 * @see sem_wait()
 */
int PLEXPORT
sem_post(sem_t *sem)
{
	if (sem == NULL) {
		errno = EINVAL;
		return -1;
	}

	platomic_inc(&((*sem)->sem_value));
	if ((*sem)->sem_wcount > 0) {
		if (plsys_tsignal(&((*sem)->sem_value), 0) < 0) {
			return -1;
		}
	}

	return  0;
}


/**
 * Decrement (lock) a semaphore.
 * In blocking mode this function is a cancellation point.
 *
 * @param sem (IN) Semaphore to lock.
 *
 * @param block (IN) If non-zero the thread is put to sleep until
 * the lock is done otherwise is will just try to lock it and return.
 *
 * @return <code>0</code> on success;
 * if <code>block</code> is <code>0</code> and the semaphore could'nt
 * be locked <code>-1</code> is returned and <code>errno</code> is set
 * to <code>EAGAIN</code>;
 * on error <code>-1</code> is returned and <code>errno</code> is set.
 * 
 * @see sem_post()
 */
static int
_sem_wait(sem_t *sem, int block)
{
	PLSTATUS st;
	long value;

	if (sem == NULL) {
		errno = EINVAL;
		return -1;
	}

	value = platomic_dec_positive(&(*sem)->sem_value);
	if (value > 0) {
		return 0;
	}

	if (!block) {
		errno = EAGAIN;
		return -1;
	}

	plenter_cancel_point();

	platomic_inc(&(*sem)->sem_wcount);

	while (1) {
		st = plsys_twait(&(*sem)->sem_value, value, 0);
		if (!(st == PLSTATUS_FAILED && errno == EWOULDBLOCK)) {
			break;
		}
		value = platomic_dec_positive(&(*sem)->sem_value);
		if (value > 0) {
			st = PLSTATUS_SUCCESS;
			break;
		}
	}

	platomic_dec(&(*sem)->sem_wcount);

	plleave_cancel_point();

	return ((st == PLSTATUS_SUCCESS) ? 0 : -1);
}


/**
 * Try to decrement (lock) a semaphore.
 *
 * @param sem (IN) Semaphore to try to lock.
 *
 * @return <code>0</code> on success;
 * if the semaphore could'nt be locked <code>-1</code> is returned
 * and <code>errno</code> is set to <code>EAGAIN</code>;
 * on error <code>-1</code> is returned and <code>errno</code>
 * indicates the error.
 * 
 * @see _sem_wait()
 * @see sem_post()
 * @see sem_wait()
 */
int PLEXPORT
sem_trywait(sem_t *sem)
{
	return _sem_wait(sem, 0);
}


/**
 * Decrement (lock) a semaphore.
 * If the semaphore cannot be locked the thread is put to sleep
 * until the operation is done.
 * <br>
 * This function is a cancellation point.
 *
 * @param sem (IN) Semaphore to lock.
 *
 * @return <code>0</code> on success; <code>-1</code> on error
 * (<code>errno</code> is set properly).
 *
 * @see _sem_wait()
 * @see sem_post()
 * @see sem_trywait()
 */
int PLEXPORT
sem_wait(sem_t *sem)
{
	return _sem_wait(sem, 1);
}


