/**
 * @file src/lib/pthread/signal.c
 * Peludo Threads - Signals.
 */
#include <plc.h>
#include <pllog.h>

#include <pthread.h>
#include <signal.h>

#include "pthread_p.h"


/**
 * Send a signal to a thread.
 *
 * @param th (IN) Thread to signal;
 *
 * @param signo (IN) Signal number.
 *
 * @return On success <code>0</code>; otherwise the error number.
 */
int PLEXPORT
pthread_kill(pthread_t th, int signo)
{
	PLSTATUS st; 
	long tid;

	if (th == NULL) {
		return ESRCH;
	}

	tid = plsys_gettid();

	plock_spvalock(&th->thread_lock, tid);

	/*
	 * Don't signal dead threads.
	 */
	if (platomic_iex(&th->thread_state, PTHREAD_STATE_DEAD,
	    PTHREAD_STATE_DEAD) == PTHREAD_STATE_DEAD) {
		pldebug("%s: Trying to signal a defunct thread, "
		    "singal=%d, thread=%p, pid=%d, tid=%d. Ignoring\n",
		    __FUNCTION__,
		    signo,
		    th,
		    th->thread_pid,
		    th->thread_tid);

		st = PLSTATUS_SUCCESS;

	} else {
		pldebug("%s: Sending signal=%d to thread=%p, pid=%d, tid=%d\n",
		    __FUNCTION__,
		    signo,
		    th,
		    th->thread_pid,
		    th->thread_tid);

		st = plsys_tkill(th->thread_tid, signo);

		if (st != PLSTATUS_SUCCESS) {
			pldebug("%s: Failed. Signal=%d, thread=%p, pid=%d, "
			    "tid=%d\n",
			    __FUNCTION__,
			    signo,
			    th,
			    th->thread_pid,
			    th->thread_tid);
		}
	}

	plock_unvalock(&th->thread_lock, tid);

	return ((st == PLSTATUS_SUCCESS) ? 0 : errno);
}


/**
 * Send a signal to all threads but the main thread.
 *
 * @param signo (IN) Signal number.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_kill()
 */
int PLEXPORT
pthread_killall_np(int signo)
{
	long tid, err;
	pthread_t th;
	sigset_t set;

	err = 0;

	tid = plsys_gettid();

	plsigblock(&set);
	plock_spvalock(&_pthread_list_lock, tid);

	PLIST_FOREACH(&_pthread_list, th, thread_list) {
		/*
		 * Don't signal the main thread.
		 */
		if (th == _pthread_main) {
			continue;
		}

		err |= pthread_kill(th, signo);
	}

	plock_unvalock(&_pthread_list_lock, tid);
	plsigunblock(&set);

	return err;
}


/**
 * Change the signal mask of the calling thread.
 *
 * @param how (IN) If <code>SIG_SETMASK</code> the signal set is set
 * to newmask; if <code>SIG_BLOCK</code> the signals inside the newmask
 * are added to the current mask; if <code>SIG_UNBLOCK</code> the signals
 * inside the newmask are removed from the current mask;
 *
 * @param newmask (IN) Set of signals;
 *
 * @param oldmask (OUT) If not <code>NULL</code> it is updated with the
 * current mask.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see plsys_tsigmask()
 */
int PLEXPORT
pthread_sigmask(int how, const sigset_t *newmask, sigset_t *oldmask)
{
	return ((plsys_tsigmask(how, newmask, oldmask) == PLSTATUS_SUCCESS)
	    ? 0 : errno);
}


