/**
 * @file src/lib/pthread/tls.c
 * Peludo Threads - TLS (Thread Local Storage).
 */
#include <stdlib.h>
#include <string.h>

#include <pthread.h>

#include "pthread_p.h"


/**
 * Thread Local Storage Area.
 *
 * @see pthread_key_t
 */
typedef struct _TLS {
	PLIST		 tls_list;	/* First member	*/

	pthread_key_t	 tls_key;	/* TLS key	*/
	void		*tls_data;	/* TLS data	*/
} TLS;


/**
 * Local Storage Key Entry.
 */
typedef struct _KEY_ENTRY {
	int	key_used;
	void	(*key_destructor)(void *);
} KEY_ENTRY;


/**
 * Array of Local Storage Keys.
 *
 * Due to the peludo kernel coding constraints we can't declare
 * an empty array, it should be created dynamically.
 * Although, a normal kernel does not embed the pthread library.
 *
 * @see KEY_ENTRY
 */
static KEY_ENTRY *_tls_keys = NULL;
static PLOCK _tls_keys_lock = PLOCK_INITIALIZED;


/**
 * Get the current thread's TLS Area associated to a specified key.
 * If the storage does not exists it is created.
 *
 * @param key (IN) Key associated to the desired TLS area.
 *
 * @return On success a pointer to the TLS area is returned;
 * on error (out of memory) <code>NULL</code> is returned.
 *
 * @see pthread_key_t
 * @see TLS
 */
static TLS *
_tls_lookup(pthread_key_t key)
{
	pthread_t self;
	TLS *tls;

	self = pthread_self();
	if (self == NULL) {
		/*
		 * Not a peludo thread.
		 */
		plsys_exit(-1);
		/* NOTREACHED */
	}

	PLIST_FOREACH(&self->thread_tlslist, tls, tls_list) {
		if (tls->tls_key == key) {
			return tls;
		}
	}

	pldebug("%s: Creating TLS Area for key %d\n", __FUNCTION__, key);

	tls = (TLS *)malloc(sizeof(TLS));
	if (tls == NULL) {
		return NULL;
	}

	PLIST_INIT(&tls->tls_list);
	tls->tls_key = key;
	tls->tls_data = NULL;

	plist_add_tail((PLIST *)&self->thread_tlslist, (PLIST *)tls);

	return tls;
}


/**
 * Initialize the Thread Local Storage system.
 * This function is called by <code>pthread_init_np()</code>
 * during the initialization of the pthread library.
 *
 * @return A Peludo Status code.
 *
 * @see PLSTATUS
 * @see tls_release()
 */
PLSTATUS
tls_init()
{
	PLOCK_INIT_UNLOCKED(&_tls_keys_lock);

	_tls_keys = (KEY_ENTRY *)malloc(sizeof(KEY_ENTRY) * PTHREAD_KEYS_MAX);
	if (_tls_keys == NULL) {
		return PLSTATUS_OUT_OF_MEMORY;
	}

	bzero((void *)_tls_keys, sizeof(KEY_ENTRY) * PTHREAD_KEYS_MAX);

	return PLSTATUS_SUCCESS;
}


/**
 * Release the Thread Local Storage.
 * This function is called by <code>pthread_release_np()</code>
 * when the pthread library is being unloaded.
 *
 * @see _tls_keys
 * @see tls_init()
 */
void
tls_release()
{
	if (_tls_keys != NULL) {
		free(_tls_keys);
	}
}


/**
 * Release all the TLS areas of a given thread.
 * This function is called by <code>pthread_exit()</code>.
 *
 * @param th (IN) Thread whose TLS should be released.
 *
 * @see pthread_t
 * @see pthread_getspecific()
 * @see pthread_setspecific()
 */
void
tls_destructor(pthread_t th)
{
	KEY_ENTRY *entry;
	TLS *tls;

	pldebug("%s: Releasing TLS of thread=%p, tid=%d\n",
	    __FUNCTION__,
	    th,
	    th->thread_tid);

	/*
	 * Don't lock the context because this function is supposed to
	 * be called by the thread itself at exit time or when it was
	 * already terminated.
	 */

	while (!plist_empty(&th->thread_tlslist)) {

		tls = PLIST_FIRST(&th->thread_tlslist, TLS *);

		entry = _tls_keys + tls->tls_key;

		if (entry->key_destructor != NULL) {
			entry->key_destructor(tls->tls_data);
		}

		plist_del((PLIST *)tls);

		free(tls);
	}
}


/**
 * Create a TLS key.
 *
 * @param key (OUT) Updated with the key;
 *
 * @param destructor (IN) Function to be called by the system to release
 * the data associated to the key when the thread exits; if this parameter
 * is <code>NULL</code> nothing will be released.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_key_t
 * @see pthread_key_delete()
 */
int PLEXPORT
pthread_key_create(pthread_key_t *key, void (*destructor)(void *))
{
	KEY_ENTRY *entry;
	pthread_key_t k;
	sigset_t set;

	if (key == NULL) {
		return EINVAL;
	}

	plsigblock(&set);
	plock_splock(&_tls_keys_lock);

	for (k = 0, entry = _tls_keys; k < PTHREAD_KEYS_MAX; k++, entry++) {
		if (!entry->key_used) {
			entry->key_used = 1;
			entry->key_destructor = destructor;
			plock_unlock(&_tls_keys_lock);
			plsigunblock(&set);
			*key = k;
			return 0;
		}
	}

	plock_unlock(&_tls_keys_lock);
	plsigunblock(&set);

	/*
	 * All keys are used, try again later.
	 */
	return EAGAIN;
}


/**
 * Delete a TLS key.
 * Before deleting a key you should ensure
 * that the data associated to it is released.
 *
 * @param key (IN) Key to delete.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_key_t
 * @see pthread_key_create()
 */
int PLEXPORT
pthread_key_delete(pthread_key_t key)
{
	sigset_t set;

	if (key >= 0 && key < PTHREAD_KEYS_MAX) {

		plsigblock(&set);
		plock_splock(&_tls_keys_lock);

		(_tls_keys + key)->key_used = 0;
		(_tls_keys + key)->key_destructor = NULL;

		plock_unlock(&_tls_keys_lock);
		plsigunblock(&set);

		return 0;
	}

	return EINVAL;
}


/**
 * Associate some data to a specified key.
 *
 * @param key (IN) Key;
 *
 * @param data (IN) Data to be associated to the key.
 *
 * @return On success <code>0</code>; otherwise the error number.
 *
 * @see pthread_key_t
 * @see _tls_lookup()
 * @see pthread_getspecific()
 */
int PLEXPORT
pthread_setspecific(pthread_key_t key, const void *data)
{
	TLS *tls;

	if (key >= 0 && key < PTHREAD_KEYS_MAX) {
		tls = _tls_lookup(key);
		if (tls == NULL) {
			return ENOMEM;
		}

		tls->tls_data = (void *)data;
		return 0;
	}

	pldebug("%s: Invalid key <%d>\n", __FUNCTION__, key);
	return EINVAL;
}


/**
 * Get the data associated to a specified key.
 *
 * @param key (IN) Key.
 *
 * @return If the key is invalid <code>NULL</code> is returned;
 * otherwise a pointer to the desired data is returned
 * (it could be <code>NULL</code>).
 *
 * @see pthread_key_t
 * @see _tls_lookup()
 * @see pthread_setspecific()
 */
void PLEXPORT *
pthread_getspecific(pthread_key_t key)
{
	TLS *tls;

	if (key < 0 || key >= PTHREAD_KEYS_MAX) {
		pldebug("%s: Invalid key <%d>\n", __FUNCTION__, key);
		return NULL;
	}

	tls = _tls_lookup(key);

	return ((tls == NULL) ? NULL : tls->tls_data);
}


